from .main import CycleController
from .conn import Timing, PSCycler, LinacPSCycler, PSCyclerFBP, FOFBPSCycler
from .util import get_psnames, get_sections, get_trigger_by_psname, \
    TRIGGER_NAMES
from .bo_cycle_data import \
    DEFAULT_RAMP_AMPLITUDE, DEFAULT_RAMP_DURATION, DEFAULT_RAMP_NRCYCLES,\
    DEFAULT_RAMP_TOTDURATION, BASE_RAMP_CURVE_ORIG, \
    bo_get_default_waveform, bo_generate_base_waveform
from .li_cycle_data import li_get_default_waveform
from .fc_cycle_data import fc_get_default_waveform



"""Define Slit device high level PVs database.

This is not a primary source database. Primary sources can be found in:
 - Pos. Difference Control IOC: https://github.com/lnls-dig/diff-ctrl-epics-ioc
 - Motors IOC: https://github.com/lnls-dig/galil-dmc30017-epics-ioc
"""


def get_slit_database():
    """Return Slit device database."""
    pvs_database = {
        # Low level devides prefixes
        'NegativeMotionCtrl-Cte': {'type': 'string'},
        'PositiveMotionCtrl-Cte': {'type': 'string'},

        # PVs to control slit edges positions
        'NegativeEdgePos-SP': {'type': 'float', 'unit': 'mm', 'prec': 5,
                               'hihi': 1000000, 'high': 1000000,
                               'hilim': 1000000, 'lolim': -1000000,
                               'low': -1000000, 'lolo': -1000000},
        'NegativeEdgePos-RB': {'type': 'float', 'unit': 'mm', 'prec': 5,
                               'hihi': 1000000, 'high': 1000000,
                               'hilim': 1000000, 'lolim': -1000000,
                               'low': -1000000, 'lolo': -1000000},
        'PositiveEdgePos-SP': {'type': 'float', 'unit': 'mm', 'prec': 5,
                               'hihi': 1000000, 'high': 1000000,
                               'hilim': 1000000, 'lolim': -1000000,
                               'low': -1000000, 'lolo': -1000000},
        'PositiveEdgePos-RB': {'type': 'float', 'unit': 'mm', 'prec': 5,
                               'hihi': 1000000, 'high': 1000000,
                               'hilim': 1000000, 'lolim': -1000000,
                               'low': -1000000, 'lolo': -1000000},

        # Helper PVs to perform step increments in slit edges positions
        'NegativeEdgeStep-SP': {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'value': 1, 'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': -1000000,
                                'low': -1000000, 'lolo': -1000000},
        'NegativeEdgeStep-RB': {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'value': 1, 'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': -1000000,
                                'low': -1000000, 'lolo': -1000000},
        'IncNegativeEdge-Cmd': {'type': 'int', 'value': 0},
        'DecNegativeEdge-Cmd': {'type': 'int', 'value': 0},
        'PositiveEdgeStep-SP': {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'value': 1, 'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': -1000000,
                                'low': -1000000, 'lolo': -1000000},
        'PositiveEdgeStep-RB': {'type': 'float', 'unit': 'mm', 'prec': 5,
                                'value': 1, 'hihi': 1000000, 'high': 1000000,
                                'hilim': 1000000, 'lolim': -1000000,
                                'low': -1000000, 'lolo': -1000000},
        'IncPositiveEdge-Cmd': {'type': 'int', 'value': 0},
        'DecPositiveEdge-Cmd': {'type': 'int', 'value': 0},

        # Alternative PVs to control slit center and width
        'Center-SP': {'type': 'float', 'unit': 'mm', 'prec': 5},
        'Center-RB': {'type': 'float', 'unit': 'mm', 'prec': 5},
        'Width-SP':  {'type': 'float', 'unit': 'mm', 'prec': 5,
                      'hihi': 500, 'high': 500, 'hilim': 500,
                      'lolim': 0, 'low': 0, 'lolo': 0},
        'Width-RB':  {'type': 'float', 'unit': 'mm', 'prec': 5,
                      'hihi': 500, 'high': 500, 'hilim': 500,
                      'lolim': 0, 'low': 0, 'lolo': 0},

        # Helper PVs to perform step increments in slit center and width
        'CenterStep-SP': {'type': 'float', 'unit': 'mm', 'prec': 5,
                          'value': 1, 'hihi': 1000000, 'high': 1000000,
                          'hilim': 1000000, 'lolim': -1000000,
                          'low': -1000000, 'lolo': -1000000},
        'CenterStep-RB': {'type': 'float', 'unit': 'mm', 'prec': 5,
                          'value': 1, 'hihi': 1000000, 'high': 1000000,
                          'hilim': 1000000, 'lolim': -1000000,
                          'low': -1000000, 'lolo': -1000000},
        'IncCenter-Cmd': {'type': 'int', 'value': 0},
        'DecCenter-Cmd': {'type': 'int', 'value': 0},
        'WidthStep-SP':  {'type': 'float', 'unit': 'mm', 'prec': 5,
                          'value': 1, 'hihi': 1000000, 'high': 1000000,
                          'hilim': 1000000, 'lolim': -1000000,
                          'low': -1000000, 'lolo': -1000000},
        'WidthStep-RB':  {'type': 'float', 'unit': 'mm', 'prec': 5,
                          'value': 1, 'hihi': 1000000, 'high': 1000000,
                          'hilim': 1000000, 'lolim': -1000000,
                          'low': -1000000, 'lolo': -1000000},
        'IncWidth-Cmd':  {'type': 'int', 'value': 0},
        'DecWidth-Cmd':  {'type': 'int', 'value': 0},

        # Command to edges motors return to Home position
        'Home-Cmd': {'type': 'int', 'value': 0},

        # PVs indicating movement status of slit edges motors
        'NegativeDoneMov-Mon': {'type': 'int', 'value': 1},
        'PositiveDoneMov-Mon': {'type': 'int', 'value': 1},

        # PVs to perform commands to force slit edges positioning
        'ForceNegativeEdgePos-Cmd': {'type': 'float', 'value': 0},
        'ForcePositiveEdgePos-Cmd': {'type': 'float', 'value': 0},
        'ForceComplete-Mon':        {'type': 'int', 'value': 1},

        # Slit edges position limits PVs
        'LowOuterLim-SP':  {'type': 'float', 'unit': 'mm', 'prec': 5,
                            'value': -1000, 'hihi': 1000000, 'high': 1000000,
                            'hilim': 1000000, 'lolim': -1000000,
                            'low': -1000000, 'lolo': -1000000},
        'LowOuterLim-RB':  {'type': 'float', 'unit': 'mm', 'prec': 5,
                            'value': -1000, 'hihi': 1000000, 'high': 1000000,
                            'hilim': 1000000, 'lolim': -1000000,
                            'low': -1000000, 'lolo': -1000000},
        'HighOuterLim-SP': {'type': 'float', 'unit': 'mm', 'prec': 5,
                            'value': 1000, 'hihi': 1000000, 'high': 1000000,
                            'hilim': 1000000, 'lolim': -1000000,
                            'low': -1000000, 'lolo': -1000000},
        'HighOuterLim-RB': {'type': 'float', 'unit': 'mm', 'prec': 5,
                            'value': 1000, 'hihi': 1000000, 'high': 1000000,
                            'hilim': 1000000, 'lolim': -1000000,
                            'low': -1000000, 'lolo': -1000000}
    }
    return pvs_database



"""Main module of Machine Shift Application."""
import time as _time
import logging as _log
import epics as _epics
import numpy as _np

from ..util import update_bit as _updt_bit, get_bit as _get_bit
from ..namesys import SiriusPVName as _PVName
from ..epics import PV as _PV
from ..callbacks import Callback as _Callback
from ..clientarch import Time as _Time

from ..search import PSSearch as _PSSearch, HLTimeSearch as _HLTimeSearch
from ..diagsys.lidiag.csdev import Const as _LIDiagConst, ETypes as _LIDiagEnum
from ..diagsys.psdiag.csdev import ETypes as _PSDiagEnum
from ..diagsys.rfdiag.csdev import Const as _RFDiagConst
from ..devices import InjSysStandbyHandler, EVG, EGun, CurrInfoSI, \
    PowerSupplyPU, RFKillBeam, InjSysPUModeHandler, Trigger, HLTiming

from .csdev import Const as _Const, ETypes as _ETypes, \
    get_injctrl_propty_database as _get_database, \
    get_status_labels as _get_sts_lbls
from .bias_feedback import BiasFeedback as _BiasFeedback


class App(_Callback):
    """Main application for handling machine shift."""

    SCAN_FREQUENCY = 1  # [Hz]

    def __init__(self):
        """Class constructor."""
        super().__init__()
        self._pvs_database = _get_database()

        self._mode = _Const.InjMode.Decay
        self._type = _Const.InjType.MultiBunch
        self._type_mon = _Const.InjTypeMon.Undefined
        self._pumode = _Const.PUMode.Accumulation
        self._pumode_mon = _Const.PUModeMon.Undefined
        self._p2w = {
            'Type': {
                'watcher': None,
                'status': _Const.IdleRunning.Idle
            },
            'FilaOpCurr': {
                'watcher': None,
                'status': _Const.IdleRunning.Idle
            },
            'HVOpVolt': {
                'watcher': None,
                'status': _Const.IdleRunning.Idle
            },
            'PUMode': {
                'watcher': None,
                'status': _Const.IdleRunning.Idle
            },
        }
        self._thread_watdev = None
        self._target_current = 100.0
        self._bucketlist_start = 1
        self._bucketlist_stop = 864
        self._bucketlist_step = 29
        self._isinj_delay = 0
        self._isinj_duration = 300

        self._accum_state_sts = _Const.AccumSts.Off
        self._accum_period = 5  # [s]

        self._topup_state_sts = _Const.TopUpSts.Off
        self._topup_period = 3*60  # [s]
        self._topup_headstarttime = 2.43  # [s]
        self._topup_pustandbyenbl = _Const.DsblEnbl.Dsbl
        self._topup_puwarmuptime = 30
        self._aspu_standby_state = None
        self._topup_liwarmupenbl = _Const.DsblEnbl.Enbl
        self._topup_liwarmuptime = 30
        self._liti_warmup_state = None
        self._topup_bopsstandbyenbl = _Const.DsblEnbl.Dsbl
        self._topup_bopswarmuptime = 10
        self._bops_standby_state = None
        self._topup_borfstandbyenbl = _Const.DsblEnbl.Dsbl
        self._topup_borfwarmuptime = 10
        self._borf_standby_state = None
        now = _Time.now().timestamp()
        self._topup_next = now - (now % (24*60*60)) + 3*60*60
        self._topup_nrpulses = 1
        self._topup_job = None
        self._accum_job = None
        self._abort = False
        self._setting_mode = False

        self._rfkillbeam_mon = _Const.RFKillBeamMon.Idle

        self._thread_autostop = None

        secs = ['LI', 'TB', 'BO', 'TS', 'SI', 'AS']
        self._status = {
            s: self._pvs_database['DiagStatus'+s+'-Mon']['value']
            for s in secs}
        self._status['All'] = self._pvs_database['DiagStatus-Mon']['value']
        self._status_problems = set()
        self._injstatus = self._pvs_database['InjStatus-Mon']['value']

        # use pyepics recommendations for threading
        _epics.ca.use_initial_context()

        # auxiliary diagnosis pvs
        self._pvs_diag = dict()
        for sec in secs:
            self._pvs_diag[sec] = dict()
            self._pvs_diag[sec]['TI'] = {
                n: _PV(n+':Status-Mon', connection_timeout=0.05)
                for n in _HLTimeSearch.get_hl_triggers(filters={'sec': sec})}

            if sec == 'AS':
                continue

            self._pvs_diag[sec]['PS'] = {
                n: _PV(n+':DiagStatus-Mon', connection_timeout=0.05)
                for n in _PSSearch.get_psnames(
                    {'sec': sec, 'dis': 'PS', 'dev': '(B|Q.*|S.*|CH|CV)'})}

            if sec != 'LI':
                punames = _PSSearch.get_psnames({
                    'sec': sec, 'dis': 'PU', 'dev': '.*(Kckr|Sept)',
                    'propty_name': '(?!:CCoil).*'})
                if sec == 'SI':
                    punames.remove('SI-01SA:PU-InjDpKckr')
                self._pvs_diag[sec]['PU'] = {
                    n: _PV(n+':DiagStatus-Mon', connection_timeout=0.05)
                    for n in punames}
            else:
                self._pvs_diag[sec]['PU'] = {
                    n: _PV(n+':DiagStatus-Mon', connection_timeout=0.05)
                    for n in ['LI-01:PU-Modltr-1', 'LI-01:PU-Modltr-2']}

                self._pvs_diag[sec]['RF'] = {
                    n: _PV(n+':DiagStatus-Mon', connection_timeout=0.05)
                    for n in _LIDiagConst.DEV_2_LINAME if 'RF' in n}

                self._pvs_diag[sec]['Egun'] = {
                    n: _PV(n+':DiagStatus-Mon', connection_timeout=0.05)
                    for n in _LIDiagConst.DEV_2_LINAME if 'EG' in n}

            if sec in ['BO', 'SI']:
                self._pvs_diag[sec]['RF'] = {
                    n: _PV(n+':DiagStatus-Mon', connection_timeout=0.05)
                    for n in _RFDiagConst.ALL_DEVICES if n.startswith(sec)}

        # auxiliary injsys PVs
        self._pvs_injsys = dict()
        punames = _PSSearch.get_psnames({
            'sec': '(SI|TS)', 'dev': 'Inj(Sept.*|NLKckr)',
            'propty_name': '(?!:CCoil).*'})
        for pun in punames:
            pv_pulse = _PV(pun+':Pulse-Sts', connection_timeout=0.05)
            pv_trigg = _PV(
                pun.replace('PU', 'TI')+':State-Sts', connection_timeout=0.05)
            self._pvs_injsys[pun] = [pv_pulse, pv_trigg]
        ffname = _HLTimeSearch.get_hl_triggers({'sec': 'SI', 'idx': 'FF.*'})[0]
        self._pvs_injsys[ffname] = [
            _PV(ffname+':State-Sts', connection_timeout=0.05), ]

        # auxiliary devices
        self.egun_dev = EGun(
            print_log=False, callback=self._update_dev_status)
        self._init_egun = False
        self.egun_dev.trigps.pv_object('enable').add_callback(
            self._callback_watch_eguntrig)

        self._pumode_dev = InjSysPUModeHandler(
            print_log=False, callback=self._update_dev_status)

        self._evg_dev = EVG()
        self._init_injevt = False
        self._evg_dev.pv_object('InjectionEvt-Sel').add_callback(
            self._callback_watch_injectionevt)
        self._evg_dev.set_auto_monitor('TotalInjCount-Mon', True)
        self._evg_dev.pv_object('TotalInjCount-Mon').add_callback(
            self._callback_is_injecting)

        self._injsys_dev = InjSysStandbyHandler()

        self.currinfo_dev = CurrInfoSI()
        self.currinfo_dev.set_auto_monitor('Current-Mon', True)
        curr_pvo = self.currinfo_dev.pv_object('Current-Mon')
        curr_pvo.add_callback(self._callback_autostop)
        curr_pvo.connection_callbacks.append(self._callback_conn_autostop)

        self._pu_names = _PSSearch.get_psnames({
            'dis': 'PU', 'dev': '.*(InjKckr|EjeKckr|InjNLKckr|Sept)',
            'propty_name': '(?!:CCoil).*'})
        self._pu_devs = [PowerSupplyPU(pun) for pun in self._pu_names]
        self._pu_refvolt = list()
        for dev in self._pu_devs:
            pvo = dev.pv_object('Voltage-SP')
            self._pu_refvolt.append(pvo.value)
            pvo.add_callback(self._callback_update_pu_refvolt)

        self._rfkillbeam = RFKillBeam()

        self._li_trig_names = _HLTimeSearch.get_hl_triggers(
            {'sec': 'LI', 'dev': '(Mod|LLRF|SSAmp|Osc)'})
        self._li_trig_devs = [Trigger(tin) for tin in self._li_trig_names]

        self._bops_trig_names = _HLTimeSearch.get_hl_triggers(
            {'sec': 'BO', 'dev': 'Mags'})
        self._bops_trig_devs = [Trigger(tin) for tin in self._bops_trig_names]

        self._borf_trig_names = _HLTimeSearch.get_hl_triggers(
            {'sec': 'BO', 'dev': 'LLRF', 'idx': 'Rmp'})
        self._borf_trig_devs = [Trigger(tin) for tin in self._borf_trig_names]

        self._hlti_dev = HLTiming()

        # pvname to write method map
        self.map_pv2write = {
            'Mode-Sel': self.set_mode,
            'Type-Sel': self.set_type,
            'SglBunBiasVolt-SP': self.set_sglbunbiasvolt,
            'MultBunBiasVolt-SP': self.set_multbunbiasvolt,
            'FilaOpCurr-SP': self.set_filaopcurr,
            'HVOpVolt-SP': self.set_hvopvolt,
            'PUMode-Sel': self.set_pumode,
            'PUModeDeltaPosAng-SP': self.set_pumode_delta_posang,
            'PUModeDpKckrDlyRef-SP': self.set_pumode_dpkckr_dlyref,
            'PUModeDpKckrKick-SP': self.set_pumode_dpkckr_kick,
            'TargetCurrent-SP': self.set_target_current,
            'BucketListStart-SP': self.set_bucketlist_start,
            'BucketListStop-SP': self.set_bucketlist_stop,
            'BucketListStep-SP': self.set_bucketlist_step,
            'IsInjDelay-SP': self.set_isinj_delay,
            'IsInjDuration-SP': self.set_isinj_duration,
            'AccumState-Sel': self.set_accum_state,
            'AccumPeriod-SP': self.set_accum_period,
            'TopUpState-Sel': self.set_topup_state,
            'TopUpPeriod-SP': self.set_topup_period,
            'TopUpHeadStartTime-SP': self.set_topup_headstarttime,
            'TopUpPUStandbyEnbl-Sel': self.set_topup_pustandbyenbl,
            'TopUpPUWarmUpTime-SP': self.set_topup_puwarmuptime,
            'TopUpLIWarmUpEnbl-Sel': self.set_topup_liwarmupenbl,
            'TopUpLIWarmUpTime-SP': self.set_topup_liwarmuptime,
            'TopUpBOPSStandbyEnbl-Sel': self.set_topup_bopsstandbyenbl,
            'TopUpBOPSWarmUpTime-SP': self.set_topup_bopswarmuptime,
            'TopUpBORFStandbyEnbl-Sel': self.set_topup_borfstandbyenbl,
            'TopUpBORFWarmUpTime-SP': self.set_topup_borfwarmuptime,
            'TopUpNrPulses-SP': self.set_topup_nrpulses,
            'InjSysTurnOn-Cmd': self.cmd_injsys_turn_on,
            'InjSysTurnOff-Cmd': self.cmd_injsys_turn_off,
            'InjSysTurnOnOrder-SP': self.set_injsys_on_order,
            'InjSysTurnOffOrder-SP': self.set_injsys_off_order,
            'RFKillBeam-Cmd': self.cmd_rfkillbeam,
        }

        # status scanning
        self.quit = False
        self.scanning = False
        self.thread_check_diagstatus = _epics.ca.CAThread(
            target=self._update_diagstatus, daemon=True)
        self.thread_check_diagstatus.start()
        self.thread_check_injstatus = _epics.ca.CAThread(
            target=self._update_injstatus, daemon=True)
        self.thread_check_injstatus.start()

        # initialize default operation values with implemented values
        self.egun_dev.wait_for_connection()
        biasvolt = self.egun_dev.bias.voltage
        if biasvolt is None:
            self._sglbunbiasvolt = _Const.BIAS_SINGLE_BUNCH
            self._multbunbiasvolt = _Const.BIAS_MULTI_BUNCH
        else:
            self._sglbunbiasvolt = biasvolt
            self._multbunbiasvolt = biasvolt
            self.egun_dev.single_bunch_bias_voltage = biasvolt
            self.egun_dev.multi_bunch_bias_voltage = biasvolt
        filacurr = self.egun_dev.fila.current
        if filacurr is None:
            self._filaopcurr = _Const.FILACURR_OPVALUE
        else:
            self._filaopcurr = filacurr
            self.egun_dev.fila_current_opvalue = filacurr
        hvvolt = self.egun_dev.hvps.voltage
        if hvvolt is None:
            self._hvopvolt = _Const.HV_OPVALUE
        else:
            self._hvopvolt = hvvolt
            self.egun_dev.high_voltage_opvalue = hvvolt

        # Create object to make bias feedback:
        self._bias_feedback = _BiasFeedback(self)
        self._pvs_database.update(self._bias_feedback.database)
        for prop, write in self._bias_feedback.map_pv2write.items():
            pvname = _Const.BIASFB_PROPTY_PREFIX + prop
            self.map_pv2write[pvname] = write

    def init_database(self):
        """Set initial PV values."""
        pvn2vals = {
            'Mode-Sel': self._mode,
            'Mode-Sts': self._mode,
            'TypeCmdSts-Mon': self._p2w['Type']['status'],
            'SglBunBiasVolt-SP': self._sglbunbiasvolt,
            'SglBunBiasVolt-RB': self._sglbunbiasvolt,
            'MultBunBiasVolt-SP': self._multbunbiasvolt,
            'MultBunBiasVolt-RB': self._multbunbiasvolt,
            'BiasVoltCmdSts-Mon': _Const.IdleRunning.Idle,
            'FilaOpCurr-SP': self._filaopcurr,
            'FilaOpCurr-RB': self._filaopcurr,
            'FilaOpCurrCmdSts-Mon': self._p2w['FilaOpCurr']['status'],
            'HVOpVolt-SP': self._hvopvolt,
            'HVOpVolt-RB': self._hvopvolt,
            'HVOpVoltCmdSts-Mon': self._p2w['HVOpVolt']['status'],
            'PUModeDeltaPosAng-SP': self._pumode_dev.delta_posang,
            'PUModeDeltaPosAng-RB': self._pumode_dev.delta_posang,
            'PUModeDpKckrDlyRef-SP': self._pumode_dev.dpkckr_dlyref,
            'PUModeDpKckrDlyRef-RB': self._pumode_dev.dpkckr_dlyref,
            'PUModeDpKckrKick-SP': self._pumode_dev.dpkckr_kick,
            'PUModeDpKckrKick-RB': self._pumode_dev.dpkckr_kick,
            'PUModeCmdSts-Mon': self._p2w['PUMode']['status'],
            'TargetCurrent-SP': self._target_current,
            'TargetCurrent-RB': self._target_current,
            'BucketListStart-SP': self._bucketlist_start,
            'BucketListStart-RB': self._bucketlist_start,
            'BucketListStop-SP': self._bucketlist_stop,
            'BucketListStop-RB': self._bucketlist_stop,
            'BucketListStep-SP': self._bucketlist_step,
            'BucketListStep-RB': self._bucketlist_step,
            'IsInjecting-Mon': _Const.IdleInjecting.Idle,
            'IsInjDelay-SP': self._isinj_delay,
            'IsInjDelay-RB': self._isinj_delay,
            'IsInjDuration-SP': self._isinj_duration,
            'IsInjDuration-RB': self._isinj_duration,
            'AccumState-Sel': _Const.OffOn.Off,
            'AccumState-Sts': self._accum_state_sts,
            'AccumPeriod-SP': self._accum_period,
            'AccumPeriod-RB': self._accum_period,
            'TopUpState-Sel': _Const.OffOn.Off,
            'TopUpState-Sts': self._topup_state_sts,
            'TopUpPeriod-SP': self._topup_period/60,
            'TopUpPeriod-RB': self._topup_period/60,
            'TopUpHeadStartTime-SP': self._topup_headstarttime,
            'TopUpHeadStartTime-RB': self._topup_headstarttime,
            'TopUpPUStandbyEnbl-Sel': self._topup_pustandbyenbl,
            'TopUpPUStandbyEnbl-Sts': self._topup_pustandbyenbl,
            'TopUpPUWarmUpTime-SP': self._topup_puwarmuptime,
            'TopUpPUWarmUpTime-RB': self._topup_puwarmuptime,
            'TopUpLIWarmUpEnbl-Sel': self._topup_liwarmupenbl,
            'TopUpLIWarmUpEnbl-Sts': self._topup_liwarmupenbl,
            'TopUpLIWarmUpTime-SP': self._topup_liwarmuptime,
            'TopUpLIWarmUpTime-RB': self._topup_liwarmuptime,
            'TopUpBOPSStandbyEnbl-Sel': self._topup_bopsstandbyenbl,
            'TopUpBOPSStandbyEnbl-Sts': self._topup_bopsstandbyenbl,
            'TopUpBOPSWarmUpTime-SP': self._topup_bopswarmuptime,
            'TopUpBOPSWarmUpTime-RB': self._topup_bopswarmuptime,
            'TopUpBORFStandbyEnbl-Sel': self._topup_bopsstandbyenbl,
            'TopUpBORFStandbyEnbl-Sts': self._topup_bopsstandbyenbl,
            'TopUpBORFWarmUpTime-SP': self._topup_bopswarmuptime,
            'TopUpBORFWarmUpTime-RB': self._topup_bopswarmuptime,
            'TopUpNextInj-Mon': self._topup_next,
            'TopUpNrPulses-SP': self._topup_nrpulses,
            'TopUpNrPulses-RB': self._topup_nrpulses,
            'InjSysCmdDone-Mon': ','.join(self._injsys_dev.done),
            'InjSysCmdSts-Mon': _Const.InjSysCmdSts.Idle,
            'RFKillBeam-Mon': _Const.RFKillBeamMon.Idle,
            'DiagStatusLI-Mon': self._status['LI'],
            'DiagStatusTB-Mon': self._status['TB'],
            'DiagStatusBO-Mon': self._status['BO'],
            'DiagStatusTS-Mon': self._status['TS'],
            'DiagStatusSI-Mon': self._status['SI'],
            'DiagStatus-Mon': self._status['AS'],
            'InjStatus-Mon': self._injstatus,
        }
        for pvn, val in pvn2vals.items():
            self.run_callbacks(pvn, val)

        self._callback_update_type(init=True)
        self.egun_dev.pulse.pv_object('multiselstatus').add_callback(
            self._callback_update_type)
        self.egun_dev.pulse.pv_object('multiswstatus').add_callback(
            self._callback_update_type)
        self.egun_dev.pulse.pv_object('singleselstatus').add_callback(
            self._callback_update_type)
        self.egun_dev.pulse.pv_object('singleswstatus').add_callback(
            self._callback_update_type)
        self.egun_dev.trigmultipre.pv_object('State-Sts').add_callback(
            self._callback_update_type)
        self.egun_dev.trigmulti.pv_object('State-Sts').add_callback(
            self._callback_update_type)
        self.egun_dev.trigsingle.pv_object('State-Sts').add_callback(
            self._callback_update_type)
        self._callback_update_pumode(init=True)
        self._pumode_dev.trigdpk.pv_object('Src-Sts').add_callback(
            self._callback_update_pumode)
        self._pumode_dev.trigdpk.pv_object('DelayRaw-RB').add_callback(
            self._callback_update_pumode)
        self._pumode_dev.pudpk.pv_object('Kick-RB').add_callback(
            self._callback_update_pumode)
        self._pumode_dev.pudpk.pv_object('PwrState-Sts').add_callback(
            self._callback_update_pumode)
        self._pumode_dev.pudpk.pv_object('Pulse-Sts').add_callback(
            self._callback_update_pumode)
        self._pumode_dev.punlk.pv_object('PwrState-Sts').add_callback(
            self._callback_update_pumode)
        self._pumode_dev.punlk.pv_object('Pulse-Sts').add_callback(
            self._callback_update_pumode)
        self._bias_feedback.init_database()
        self.run_callbacks('Log-Mon', 'Started.')

    @property
    def pvs_database(self):
        """Return pvs_database."""
        return self._pvs_database

    def process(self, interval):
        """Sleep."""
        _time.sleep(interval)

    def read(self, reason):
        """Read from IOC database."""
        value = None
        return value

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        _log.info('Write received for: %s --> %s', reason, str(value))
        if reason in self.map_pv2write.keys():
            status = self.map_pv2write[reason](value)
            _log.info('%s Write for: %s --> %s',
                      str(status).upper(), reason, str(value))
            return status
        _log.warning('PV %s does not have a set function.', reason)
        return False

    # ----- handle writes -----

    def set_mode(self, value):
        """Set injection mode."""
        if not 0 <= value < len(_ETypes.INJMODE):
            return False

        if value == self._mode:
            self.run_callbacks('Mode-Sts', self._mode)
            return True

        # stop topup and accum threads:
        if self._topup_job and self._topup_job.is_alive():
            self._setting_mode = True
            self._stop_topup_job()
            self._setting_mode = False
        if self._accum_job and self._accum_job.is_alive():
            self._setting_mode = True
            self._stop_accum_job()
            self._setting_mode = False

        if value != _Const.InjMode.Decay:
            stg = 'top-up' if value == _Const.InjMode.TopUp else 'accumulation'
            self._update_log('Configuring EVG RepeatBucketList...')
            self._evg_dev['RepeatBucketList-SP'] = 1
            self._update_log(f'...done. Ready to start {stg:s}.')

        self._mode = value
        self.run_callbacks('Mode-Sts', self._mode)
        return True

    def set_type(self, value):
        """Set injection type."""
        if not 0 <= value < len(_ETypes.INJTYPE):
            return False
        if self._mode != _Const.InjMode.Decay:
            self._update_log(
                f'ERR:InjType can only be changed in Decay mode.')
            return False
        if self._p2w['Type']['watcher'] is not None and \
                self._p2w['Type']['watcher'].is_alive():
            self._update_log('WARN:Interrupting type change command..')
            self.egun_dev.cmd_abort_chg_type()
            self._p2w['Type']['watcher'].join()

        self._type = value
        self.run_callbacks('Type-Sts', self._type)

        target = self.egun_dev.cmd_switch_to_single_bunch \
            if value == _Const.InjType.SingleBunch else \
            self.egun_dev.cmd_switch_to_multi_bunch
        self._p2w['Type']['watcher'] = _epics.ca.CAThread(
            target=target, daemon=True)
        self._p2w['Type']['watcher'].start()
        self._p2w['Type']['status'] = _Const.IdleRunning.Running
        self.run_callbacks('TypeCmdSts-Mon', self._p2w['Type']['status'])

        self._launch_watch_dev_thread()
        return True

    def set_sglbunbiasvolt(self, value):
        """Set single bunch bias voltage."""
        self._update_log('Received setpoint to SB Bias voltage.')
        self.egun_dev.single_bunch_bias_voltage = value
        self._sglbunbiasvolt = value
        self.run_callbacks('SglBunBiasVolt-RB', self._sglbunbiasvolt)

        if self._type == _Const.InjType.SingleBunch:
            _epics.ca.CAThread(
                target=self._set_egunbias, args=[value, ], daemon=True).start()
        return True

    def set_multbunbiasvolt(self, value):
        """Set multi bunch bias voltage."""
        self._update_log('Received setpoint to MB Bias voltage.')
        self.egun_dev.multi_bunch_bias_voltage = value
        self._multbunbiasvolt = value
        self.run_callbacks('MultBunBiasVolt-RB', self._multbunbiasvolt)

        if self._type == _Const.InjType.MultiBunch:
            _epics.ca.CAThread(
                target=self._set_egunbias, args=[value, ], daemon=True).start()
        return True

    def _set_egunbias(self, value):
        self.run_callbacks('BiasVoltCmdSts-Mon', _Const.IdleRunning.Running)

        self._update_log(f'Setting EGun Bias voltage to {value:.2f}V...')
        if not self.egun_dev.bias.set_voltage(value):
            self._update_log('ERR:Could not set EGun Bias voltage.')
        else:
            self._update_log(f'Set EGun Bias voltage: {value:.2f}V.')

        self.run_callbacks('BiasVoltCmdSts-Mon', _Const.IdleRunning.Idle)

    def set_filaopcurr(self, value):
        """Set filament current operation value."""
        if self._p2w['FilaOpCurr']['watcher'] is not None and \
                self._p2w['FilaOpCurr']['watcher'].is_alive():
            self._update_log('WARN:Interrupting FilaPS current ramp...')
            self.egun_dev.cmd_abort_rmp_fila()
            self._p2w['FilaOpCurr']['watcher'].join()

        self.egun_dev.fila_current_opvalue = value
        self._filaopcurr = value
        self.run_callbacks('FilaOpCurr-RB', self._filaopcurr)

        self._p2w['FilaOpCurr']['watcher'] = _epics.ca.CAThread(
            target=self.egun_dev.set_fila_current, daemon=True)
        self._p2w['FilaOpCurr']['watcher'].start()
        self._p2w['FilaOpCurr']['status'] = _Const.IdleRunning.Running
        self.run_callbacks(
            'FilaOpCurrCmdSts-Mon', self._p2w['FilaOpCurr']['status'])

        self._launch_watch_dev_thread()
        return True

    def set_hvopvolt(self, value):
        """Set high voltage operation value."""
        if self._p2w['HVOpVolt']['watcher'] is not None and \
                self._p2w['HVOpVolt']['watcher'].is_alive():
            self._update_log('WARN:Interrupting HVPS voltage ramp...')
            self.egun_dev.cmd_abort_rmp_hvps()
            self._p2w['HVOpVolt']['watcher'].join()

        self.egun_dev.high_voltage_opvalue = value
        self._hvopvolt = value
        self.run_callbacks('HVOpVolt-RB', self._hvopvolt)

        self._p2w['HVOpVolt']['watcher'] = _epics.ca.CAThread(
            target=self.egun_dev.set_hv_voltage, daemon=True)
        self._p2w['HVOpVolt']['watcher'].start()
        self._p2w['HVOpVolt']['status'] = _Const.IdleRunning.Running
        self.run_callbacks(
            'HVOpVoltCmdSts-Mon', self._p2w['HVOpVolt']['status'])

        self._launch_watch_dev_thread()
        return True

    def set_pumode(self, value):
        """Set PU mode."""
        if not 0 <= value < len(_ETypes.PUMODE):
            return False
        if self._mode != _Const.InjMode.Decay:
            topup = _Const.InjMode.TopUp
            stg = 'top-up' if self._mode == topup else 'accumulation'
            self._update_log(
                f'ERR:PUMode can only be changed in Decay mode.')
            return False
        if self._p2w['PUMode']['watcher'] is not None and \
                self._p2w['PUMode']['watcher'].is_alive():
            self._update_log('WARN:Interrupting PUMode change command')
            self._pumode_dev.cmd_abort()
            self._p2w['PUMode']['watcher'].join()

        self._pumode = value
        self.run_callbacks('PUMode-Sts', self._pumode)

        target = self._pumode_dev.cmd_switch_to_accum \
            if value == _Const.PUMode.Accumulation else \
            self._pumode_dev.cmd_switch_to_optim \
            if value == _Const.PUMode.Optimization else \
            self._pumode_dev.cmd_switch_to_onaxis
        self._p2w['PUMode']['watcher'] = _epics.ca.CAThread(
            target=target, daemon=True)
        self._p2w['PUMode']['watcher'].start()
        self._p2w['PUMode']['status'] = _Const.IdleRunning.Running
        self.run_callbacks('PUModeCmdSts-Mon', self._p2w['PUMode']['status'])

        self._launch_watch_dev_thread()
        return True

    def set_pumode_delta_posang(self, value):
        """Set PU mode delta posang."""
        self._pumode_dev.delta_posang = value
        self.run_callbacks('PUModeDeltaPosAng-RB', value)
        return True

    def set_pumode_dpkckr_dlyref(self, value):
        """Set PU mode DpKckr delay."""
        self._pumode_dev.dpkckr_dlyref = value
        self.run_callbacks('PUModeDpKckrDlyRef-RB', value)
        return True

    def set_pumode_dpkckr_kick(self, value):
        """Set PU mode DpKckr kick."""
        self._pumode_dev.dpkckr_kick = value
        self.run_callbacks('PUModeDpKckrKick-RB', value)
        return True

    def set_target_current(self, value):
        """Set the target injection current value ."""
        self._target_current = value
        self.run_callbacks('TargetCurrent-RB', self._target_current)
        self._update_log(
            'Updated target current value: {}mA.'.format(value))
        return True

    def set_bucketlist_start(self, start):
        """Set bucketlist_start."""
        if not _Const.MIN_BKT <= start <= _Const.MAX_BKT:
            return False
        stop = self._bucketlist_stop
        step = self._bucketlist_step
        if self._mode == _Const.InjMode.TopUp:
            if not self._cmd_bucketlist_fill(stop, start, step):
                return False
        self._bucketlist_start = start
        self.run_callbacks('BucketListStart-RB', start)
        return True

    def set_bucketlist_stop(self, stop):
        """Set bucketlist_stop."""
        if not _Const.MIN_BKT <= stop <= _Const.MAX_BKT:
            return False
        start = self._bucketlist_start
        step = self._bucketlist_step
        if self._mode == _Const.InjMode.TopUp:
            if not self._cmd_bucketlist_fill(stop, start, step):
                return False
        self._bucketlist_stop = stop
        self.run_callbacks('BucketListStop-RB', stop)
        return True

    def set_bucketlist_step(self, step):
        """Set bucketlist_step."""
        if not -_Const.MAX_BKT+1 <= step <= _Const.MAX_BKT-1:
            return False
        if self._mode != _Const.InjMode.Decay:
            if not self._update_bucket_list(step=step):
                return False
        else:
            start = self._bucketlist_start
            stop = self._bucketlist_stop
            if not self._cmd_bucketlist_fill(stop, start, step):
                return False
        self._bucketlist_step = step
        self.run_callbacks('BucketListStep-RB', step)
        return True

    def set_isinj_delay(self, value):
        """Set IsInjecting-Mon flag delay."""
        if not 0 <= value <= 1000:
            return False
        self._isinj_delay = value
        self._update_log(f'Changed IsInjecting-Mon flag delay to {value}ms.')
        self.run_callbacks('IsInjDelay-RB', value)
        return True

    def set_isinj_duration(self, value):
        """Set IsInjecting-Mon flag duration."""
        if not 0 <= value <= 1000:
            return False
        self._isinj_duration = value
        self._update_log(
            f'Changed IsInjecting-Mon flag duration to {value}ms.')
        self.run_callbacks('IsInjDuration-RB', value)
        return True

    def _cmd_bucketlist_fill(self, stop, start, step):
        """Set bucket list PV."""
        if not self._evg_dev.connected:
            self._update_log('ERR:Could not update bucket list,')
            self._update_log('ERR:EVG is disconnected.')
            return False
        if self._evg_dev.fill_bucketlist(stop, start, step, timeout=3):
            self._update_log('Updated BucketList.')
            return True
        self._update_log('WARN:Timed out waiting for BucketList.')
        return False

    def set_topup_state(self, value):
        """Set top-up state."""
        if self._mode != _Const.InjMode.TopUp:
            return False

        if value == _Const.OffOn.On:
            self._update_log('Start received!')
            if not self._check_allok_2_inject():
                return False
            if self._topup_job is None or not self._topup_job.is_alive():
                self._launch_topup_job()
        else:
            self._update_log('Stop received!')
            if self._topup_job is not None and \
                    self._topup_job.is_alive():
                self._stop_topup_job()
        return True

    def set_accum_state(self, value):
        """Set accum state."""
        if self._mode != _Const.InjMode.Accum:
            return False

        if value == _Const.OffOn.On:
            self._update_log('Start received!')
            if not self._check_allok_2_inject():
                return False
            if self._accum_job is None or not self._accum_job.is_alive():
                self._launch_accum_job()
        else:
            self._update_log('Stop received!')
            if self._accum_job is not None and self._accum_job.is_alive():
                self._stop_accum_job()
        return True

    def set_topup_period(self, value):
        """Set top-up period [min]."""
        if not 1 <= value <= 6*60:
            return False

        sec = value * 60
        if self._topup_state_sts != _Const.TopUpSts.Off:
            now = _Time.now().timestamp()
            self._topup_next = now - (now % sec) + sec
            self.run_callbacks('TopUpNextInj-Mon', self._topup_next)

        self._topup_period = sec
        self._update_log('Changed top-up period to '+str(value)+'min.')
        self.run_callbacks('TopUpPeriod-RB', value)
        return True

    def set_accum_period(self, value):
        """Set accumulation period [s]."""
        if not 1 <= value <= 60*60:
            return False

        self._accum_period = value
        self._update_log('Changed accumulation period to '+str(value)+'s.')
        self.run_callbacks('AccumPeriod-RB', value)
        return True

    def set_topup_headstarttime(self, value):
        """Set top-up head start time [s]."""
        if not 0 <= value <= 2*60:
            return False
        self._topup_headstarttime = value
        self._update_log('Changed top-up head start time to '+str(value)+'s.')
        self.run_callbacks('TopUpHeadStartTime-RB', self._topup_headstarttime)

        minwut = _np.ceil(value+1)
        self._topup_puwarmuptime = max(minwut, self._topup_puwarmuptime)
        self._topup_liwarmuptime = max(minwut, self._topup_liwarmuptime)
        self._topup_bopswarmuptime = max(minwut, self._topup_bopswarmuptime)
        self._topup_borfwarmuptime = max(minwut, self._topup_borfwarmuptime)
        pvn2val = {
            'TopUpPUWarmUpTime-': self._topup_puwarmuptime,
            'TopUpLIWarmUpTime-': self._topup_liwarmuptime,
            'TopUpBOPSWarmUpTime-': self._topup_bopswarmuptime,
            'TopUpBORFWarmUpTime-': self._topup_borfwarmuptime,
        }
        for pvn, val in pvn2val.items():
            self.run_callbacks(pvn+'SP', val)
            self.run_callbacks(pvn+'RB', val)
        return True

    def set_topup_pustandbyenbl(self, value):
        """Enable/disable PU standby between top-up injections."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False

        if value == _Const.DsblEnbl.Dsbl:
            self._handle_aspu_standby_state(_Const.StandbyInject.Inject)
        self._topup_pustandbyenbl = value
        text = 'En' if value else 'Dis'
        self._update_log(text+'abled PU standby between injections.')
        self.run_callbacks('TopUpPUStandbyEnbl-Sts', self._topup_pustandbyenbl)
        return True

    def set_topup_puwarmuptime(self, value):
        """Set PU warm up time before top-up injections."""
        if not self._topup_headstarttime+1 <= value < 2*60:
            return False
        self._topup_puwarmuptime = value
        self.run_callbacks('TopUpPUWarmUpTime-RB', self._topup_puwarmuptime)
        return True

    def set_topup_liwarmupenbl(self, value):
        """Enable/disable LI warm up before top-up injections."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False

        if value == _Const.DsblEnbl.Dsbl:
            self._handle_liti_warmup_state(state=_Const.StandbyInject.Inject)
        self._topup_liwarmupenbl = value
        text = 'En' if value else 'Dis'
        self._update_log(text+'abled LI warm up before injections.')
        self.run_callbacks('TopUpLIWarmUpEnbl-Sts', self._topup_liwarmupenbl)
        return True

    def set_topup_liwarmuptime(self, value):
        """Set LI warm up time before top-up injections."""
        if not self._topup_headstarttime+1 <= value < 2*60:
            return False
        self._topup_liwarmuptime = value
        self.run_callbacks('TopUpLIWarmUpTime-RB', self._topup_liwarmuptime)
        return True

    def set_topup_bopsstandbyenbl(self, value):
        """Enable/disable BO PS standby between top-up injections."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False

        if value == _Const.DsblEnbl.Dsbl:
            self._handle_bops_standby_state(state=_Const.StandbyInject.Inject)
        self._topup_bopsstandbyenbl = value
        text = 'En' if value else 'Dis'
        self._update_log(text+'abled BO PS standby between injections.')
        self.run_callbacks(
            'TopUpBOPSStandbyEnbl-Sts', self._topup_bopsstandbyenbl)
        return True

    def set_topup_bopswarmuptime(self, value):
        """Set BO PS warm up time before top-up injections."""
        if not self._topup_headstarttime+1 <= value < 2*60:
            return False
        self._topup_bopswarmuptime = value
        self.run_callbacks(
            'TopUpBOPSWarmUpTime-RB', self._topup_bopswarmuptime)
        return True

    def set_topup_borfstandbyenbl(self, value):
        """Enable/disable BO RF standby between top-up injections."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False

        if value == _Const.DsblEnbl.Dsbl:
            self._handle_borf_standby_state(state=_Const.StandbyInject.Inject)
        self._topup_borfstandbyenbl = value
        text = 'En' if value else 'Dis'
        self._update_log(text+'abled BO RF standby between injections.')
        self.run_callbacks(
            'TopUpBORFStandbyEnbl-Sts', self._topup_borfstandbyenbl)
        return True

    def set_topup_borfwarmuptime(self, value):
        """Set BO RF warm up time before top-up injections."""
        if not self._topup_headstarttime+1 <= value < 2*60:
            return False
        self._topup_borfwarmuptime = value
        self.run_callbacks(
            'TopUpBORFWarmUpTime-RB', self._topup_borfwarmuptime)
        return True

    def set_topup_nrpulses(self, value):
        """Set top-up number of injection pulses."""
        if not 1 <= value <= 1000:
            return False

        self._topup_nrpulses = value
        if self._mode == _Const.InjMode.TopUp:
            if not self._update_bucket_list():
                return False
        self._update_log('Changed top-up nr.pulses to '+str(value)+'.')
        self.run_callbacks('TopUpNrPulses-RB', self._topup_nrpulses)
        return True

    def cmd_injsys_turn_on(self, value=None, wait_finish=False):
        """Set turn on Injection System."""
        run = self._injsys_dev.is_running
        if run:
            self._update_log('ERR:Still processing turn '+run+' InjSystem')
            return False

        self._update_log('Sending turn on to Inj.System...')
        self.run_callbacks(
            'InjSysCmdDone-Mon', ','.join(self._injsys_dev.done))
        self._injsys_dev.cmd_turn_on(run_in_thread=True)
        thr = _epics.ca.CAThread(
            target=self._watch_injsys, args=['on', ], daemon=True)
        thr.start()
        if wait_finish:
            thr.join()
        return True

    def cmd_injsys_turn_off(self, value=None, wait_finish=False):
        """Set turn off Injection System."""
        run = self._injsys_dev.is_running
        if run:
            self._update_log('ERR:Still processing turn '+run+' InjSystem')
            return False

        self._update_log('Sending turn off to Inj.System...')
        self.run_callbacks(
            'InjSysCmdDone-Mon', ','.join(self._injsys_dev.done))
        self._injsys_dev.cmd_turn_off(run_in_thread=True)
        thr = _epics.ca.CAThread(
            target=self._watch_injsys, args=['off', ], daemon=True)
        thr.start()
        if wait_finish:
            thr.join()
        return True

    def _watch_injsys(self, cmd, timeout=_Const.RF_RMP_TIMEOUT):
        self.run_callbacks(
            'InjSysCmdSts-Mon', getattr(_Const.InjSysCmdSts, cmd.capitalize()))
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if not self._injsys_dev.is_running:
                break
            _time.sleep(0.1)
            self.run_callbacks(
                'InjSysCmdDone-Mon', ','.join(self._injsys_dev.done))
        self.run_callbacks('InjSysCmdSts-Mon', _Const.InjSysCmdSts.Idle)
        ret = self._injsys_dev.result
        if ret is None:
            msg = 'ERR:Timed out in turn '+cmd+' Inj.System.'
        elif not ret[0]:
            self._update_log('ERR:Failed to turn '+cmd+' Inj.System.')
            msgs = [ret[1][i:i+35] for i in range(0, len(ret[1]), 35)]
            for msg in msgs:
                self._update_log('ERR:'+msg)
            self._update_log('ERR:Detail list: ')
            for item in ret[2]:
                self._update_log('ERR:'+item)
        else:
            msg = 'Turned '+cmd+' Inj.System.'
            self._update_log(msg)

    def set_injsys_on_order(self, value):
        """Set inj.sys. turn on command order."""
        new_order = value.split(',')
        if set(new_order) - set(self._injsys_dev.DEF_ON_ORDER):
            self._update_log('ERR:Invalid value for inj.sys. on order')
            return False
        self._injsys_dev.on_order = new_order
        self._update_log('Updated inj.sys. turn on command order.')
        self.run_callbacks('InjSysTurnOnOrder-RB', value)
        return True

    def set_injsys_off_order(self, value):
        """Set inj.sys. turn off command order."""
        new_order = value.split(',')
        if set(new_order) - set(self._injsys_dev.DEF_OFF_ORDER):
            self._update_log('ERR:Invalid value for inj.sys. off order')
            return False
        self._injsys_dev.off_order = new_order
        self._update_log('Updated inj.sys. turn off command order.')
        self.run_callbacks('InjSysTurnOffOrder-RB', value)
        return True

    def cmd_rfkillbeam(self, value):
        """RF Kill Beam command."""
        if self._rfkillbeam_mon == _Const.RFKillBeamMon.Kill:
            self._update_log('ERR:Still processing RFKillBeam command')
            return False

        self._update_log('Received RFKillBeam Command...')
        self._rfkillbeam_mon = _Const.RFKillBeamMon.Kill
        self.run_callbacks('RFKillBeam-Mon', self._rfkillbeam_mon)
        _epics.ca.CAThread(target=self._watch_rfkillbeam, daemon=True).start()
        return True

    def _watch_rfkillbeam(self):
        ret = self._rfkillbeam.cmd_kill_beam()
        if not ret[0]:
            msgs = [ret[1][i:i+35] for i in range(0, len(ret[1]), 35)]
            for msg in msgs:
                self._update_log('ERR:'+msg)
        else:
            self._update_log('The beam was killed by RF!')
        self._rfkillbeam_mon = _Const.RFKillBeamMon.Idle
        self.run_callbacks('RFKillBeam-Mon', self._rfkillbeam_mon)

    # --- callbacks ---

    def _callback_watch_eguntrig(self, value, **kws):
        if not self._init_egun:
            self._init_egun = True
            return
        if self._mode != _Const.InjMode.Decay:
            return
        _epics.ca.CAThread(
            target=self._watch_eguntrig, args=[value, ], daemon=True).start()

    def _watch_eguntrig(self, value, **kws):
        cmd = 'on' if value else 'off'
        _t0 = _time.time()
        while _time.time() - _t0 < 10:
            if self.egun_dev.trigps.is_on() == value:
                msg = 'Turned '+cmd+' EGun.'
                break
            _time.sleep(0.1)
        else:
            msg = 'WARN:Timed out in turn '+cmd+' Egun.'
        self._update_log(msg)

    def _callback_watch_injectionevt(self, value, **kws):
        if not self._init_injevt:
            self._init_injevt = True
            return
        if self._mode != _Const.InjMode.Decay:
            return
        _epics.ca.CAThread(
            target=self._watch_injti, args=[value, ], daemon=True).start()

    def _watch_injti(self, value, timeout=_Const.TI_INJ_TIMEOUT):
        cmd = 'on' if value else 'off'
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if self._evg_dev.injection_state == value:
                msg = 'Turned '+cmd+' InjectionEvt.'
                break
            _time.sleep(0.1)
        else:
            msg = 'WARN:Timed out in turn '+cmd+' InjectionEvt.'
        self._update_log(msg)

    def _callback_autostop(self, value, **kws):
        if self._thread_autostop is not None and \
                self._thread_autostop.is_alive():
            return
        if value is None or value < self._target_current:
            return
        self._thread_autostop = _epics.ca.CAThread(
            target=self._thread_run_autostop, args=[value, 'cb_val'])
        self._thread_autostop.start()

    def _callback_conn_autostop(self, conn, **kws):
        if self._thread_autostop is not None and \
                self._thread_autostop.is_alive():
            return
        if conn:
            return
        self._thread_autostop = _epics.ca.CAThread(
            target=self._thread_run_autostop, args=[conn, 'cb_conn'])
        self._thread_autostop.start()

    def _thread_run_autostop(self, value, cb_type):
        if self._mode != _Const.InjMode.Decay:
            return
        if not self._evg_dev['InjectionEvt-Sel']:
            return
        if not self.egun_dev.trigps.is_on():
            return
        if cb_type == 'cb_val':
            msg = 'Target current reached!'
        else:
            msg = 'ERR:Current PV disconnected.'
        self._update_log(msg)
        self._run_autostop()

    def _run_autostop(self):
        self._update_log('Running Auto Stop...')
        if self._stop_injection():
            self._update_log('Injection Auto Stop done.')
            self._update_bucket_list_autostop()
        else:
            self._update_log('ERR:Injection Auto Stop failed.')

    def _callback_update_type(self, **kws):
        if self.egun_dev.is_single_bunch:
            self._type_mon = _Const.InjTypeMon.SingleBunch
        elif self.egun_dev.is_multi_bunch:
            self._type_mon = _Const.InjTypeMon.MultiBunch
        else:
            self._type_mon = _Const.InjTypeMon.Undefined
        self.run_callbacks('Type-Mon', self._type_mon)

        if 'init' in kws and kws['init']:
            if self._type_mon != _Const.InjTypeMon.Undefined:
                self._type = _ETypes.INJTYPE.index(
                    _ETypes.INJTYPE_MON[self._type_mon])
            self.run_callbacks('Type-Sel', self._type)
            self.run_callbacks('Type-Sts', self._type)

    def _callback_update_pumode(self, **kws):
        if self._pumode_dev.is_accum:
            self._pumode_mon = _Const.PUModeMon.Accumulation
        elif self._pumode_dev.is_optim:
            self._pumode_mon = _Const.PUModeMon.Optimization
        elif self._pumode_dev.is_onaxis:
            self._pumode_mon = _Const.PUModeMon.OnAxis
        else:
            self._pumode_mon = _Const.PUModeMon.Undefined
        self.run_callbacks('PUMode-Mon', self._pumode_mon)

        if 'init' in kws and kws['init']:
            if self._pumode_mon != _Const.PUModeMon.Undefined:
                self._pumode = _ETypes.PUMODE.index(
                    _ETypes.PUMODE_MON[self._pumode_mon])
            self.run_callbacks('PUMode-Sel', self._pumode)
            self.run_callbacks('PUMode-Sts', self._pumode)

    def _callback_update_pu_refvolt(self, pvname, value, **kws):
        if value is None:
            return
        if self._aspu_standby_state == _Const.StandbyInject.Standby:
            return
        devname = _PVName(pvname).device_name
        index = self._pu_names.index(devname)
        self._pu_refvolt[index] = value

    def _callback_is_injecting(self, value, **kws):
        if value is None:
            return
        thread = _epics.ca.CAThread(
            target=self._thread_is_injecting, daemon=True)
        thread.start()

    def _thread_is_injecting(self):
        # check if any InjSI PU is pulsing
        is_injecting = False
        for pvs in self._pvs_injsys.values():
            if not all(pvo.connected for pvo in pvs):
                is_injecting = True  # assume the worst scenario
                break
            if all(pvo.value for pvo in pvs):
                is_injecting = True
                break

        if not is_injecting:
            return

        # if True, raise IsInjecting flag after {delay}ms for {duration}ms
        _time.sleep(self._isinj_delay/1000)
        self.run_callbacks('IsInjecting-Mon', _Const.IdleInjecting.Injecting)
        _time.sleep(self._isinj_duration/1000)
        self.run_callbacks('IsInjecting-Mon', _Const.IdleInjecting.Idle)

    # --- auxiliary injection methods ---

    def _check_allok_2_inject(self, show_warn=True):
        if show_warn:
            if self._status['All'] != 0:
                self._update_log('WARN:DiagStatus not ok:')
                for prob in self._status_problems:
                    self._update_log('WARN:Verify '+prob+'!')

            if self._injstatus != 0:
                self._update_log('WARN:InjStatus not ok:')
                for bit, prob in enumerate(_Const.INJ_STATUS_LABELS):
                    if _get_bit(self._injstatus, bit):
                        self._update_log('WARN:'+prob)

        if self._mode != _Const.InjMode.Decay:
            if self._evg_dev.nrpulses != 1:
                self._update_log('ERR:Aborted. RepeatBucketList must be 1.')
                return False

            if self._abort:
                self._update_log('Abort received.')
                return False
        return True

    def _start_injection(self):
        # turn on injectionevt
        self._update_log('Turning InjectionEvt on...')
        if not self._evg_dev.cmd_turn_on_injection():
            self._update_log('ERR:Could not turn on InjectionEvt.')
            return False
        self._update_log('Sent turn on to InjectionEvt.')

        # wait for injectionevt to be ready
        self._update_log('Waiting for InjectionEvt to be on...')
        _t0 = _time.time()
        while _time.time() - _t0 < _Const.TI_INJ_TIMEOUT:
            if not self._check_allok_2_inject(show_warn=False):
                self._stop_injection()
                return False
            if self._evg_dev.injection_state:
                break
            _time.sleep(0.02)
        else:
            self._update_log('ERR:Timed out waiting for InjectionEvt.')
            return False

        self._update_log('InjectionEvt is on!')
        return True

    def _wait_injection(self):
        # wait for injectionevt to be off (done)
        _t0 = _time.time()
        while _time.time() - _t0 < _Const.MAX_INJTIMEOUT:
            if not self._check_allok_2_inject(show_warn=False):
                self._stop_injection()
                return False
            if not self._evg_dev.injection_state:
                break
            _time.sleep(0.02)
        return True

    def _stop_injection(self):
        self._update_log('Turning off InjectionEvt...')
        if self._evg_dev.cmd_turn_off_injection():
            msg = 'Turned off InjectionEvt.'
        else:
            self._update_log('ERR:Failed to turn off InjectionEvt.')
            self._update_log('Turning off EGun TriggerPS...')
            if self.egun_dev.trigps.cmd_disable_trigger():
                msg = 'Turned off EGun TriggerPS.'
            else:
                msg = 'ERR:Failed to turn off EGun TriggerPS.'
        self._update_log(msg)
        return 'ERR' not in msg

    def _update_bucket_list_autostop(self):
        if not self._evg_dev.connected:
            self._update_log('ERR:Could not update bucket list,')
            self._update_log('ERR:EVG is disconnected.')
            return False
        old_bucklist = self._evg_dev.bucketlist_mon
        injcount = self._evg_dev.injection_count
        blistlen = self._evg_dev.bucketlist_len
        proll = int(injcount % blistlen)
        new_bucklist = _np.roll(old_bucklist, -1 * proll)
        return self._set_bucket_list(new_bucklist)

    def _update_bucket_list(self, step=None, nrpulses=None):
        if not self._evg_dev.connected:
            self._update_log('ERR:Could not update bucket list,')
            self._update_log('ERR:EVG is disconnected.')
            return False

        if step is None:
            step = self._bucketlist_step

        lastfilledbucket = self._evg_dev.bucketlist_mon[-1]
        if not _Const.MIN_BKT <= lastfilledbucket <= _Const.MAX_BKT:
            lastfilledbucket = 1

        nrpulses = nrpulses or self._topup_nrpulses
        bucket = _np.arange(nrpulses) + 1
        bucket *= step
        bucket += lastfilledbucket - 1
        bucket %= 864
        bucket += 1
        return self._set_bucket_list(bucket)

    def _set_bucket_list(self, value):
        self._evg_dev.bucketlist = value
        _t0 = _time.time()
        while _time.time() - _t0 < 5:
            if _np.all(self._evg_dev.bucketlist == value):
                self._update_log('Updated BucketList.')
                return True
        self._update_log('WARN:Could not update BucketList.')
        return False

    # --- auxiliary accumulation methods ---

    def _launch_accum_job(self):
        while self._abort:
            _time.sleep(0.1)
        self._update_log('Launching accumulation thread...')
        self._accum_job = _epics.ca.CAThread(
            target=self._do_accumulation, daemon=True)
        self._accum_job.start()

    def _stop_accum_job(self):
        if self._abort:
            return
        self._update_log('Stopping accumulation thread...')
        self._abort = True
        self._accum_job.join()
        self._accum_job = None
        self._update_log('Stopped accumulation thread.')
        self._abort = False

    def _do_accumulation(self):
        # update bucket list according to settings
        self._update_bucket_list(nrpulses=1)

        self._handle_aspu_standby_state(_Const.StandbyInject.Inject)
        self._handle_liti_warmup_state(_Const.StandbyInject.Inject)
        self._handle_bops_standby_state(_Const.StandbyInject.Inject)
        self._handle_borf_standby_state(_Const.StandbyInject.Inject)

        while self._mode == _Const.InjMode.Accum:
            t0_ = _time.time()
            if not self.currinfo_dev.connected:
                self._update_log('ERR:CurrInfo device disconnected.')
                break
            if self.currinfo_dev.current >= self._target_current:
                self._update_log(
                    'Target Current reached. Stopping accumulation...')
                break

            if not self._check_allok_2_inject():
                break

            self.run_callbacks('AccumState-Sts', _Const.AccumSts.TurningOn)
            if not self._start_injection():
                break

            self.run_callbacks('AccumState-Sts', _Const.AccumSts.Injecting)
            if not self._wait_injection():
                break
            self._update_bucket_list(nrpulses=1)

            dt_ = self._accum_period - (_time.time() - t0_)
            if dt_ > 0:
                self.run_callbacks('AccumState-Sts', _Const.AccumSts.Waiting)
                self._update_log('Waiting for next injection...')
                _time.sleep(dt_)

        self._handle_liti_warmup_state(_Const.StandbyInject.Standby)

        # update top-up status
        self.run_callbacks('AccumState-Sts', _Const.AccumSts.Off)
        self._update_log('Stopped accumulation loop.')
        if not self._abort or self._setting_mode:
            self.run_callbacks('AccumState-Sel', _Const.OffOn.Off)

    # --- auxiliary top-up methods ---

    def _launch_topup_job(self):
        while self._abort:
            _time.sleep(0.1)
        self._update_log('Launching top-up thread...')
        self._topup_job = _epics.ca.CAThread(
            target=self._do_topup, daemon=True)
        self._topup_job.start()

    def _stop_topup_job(self):
        if self._abort:
            return
        self._update_log('Stopping top-up thread...')
        self._abort = True
        self._topup_job.join()
        self._topup_job = None
        self._update_log('Stopped top-up thread.')
        self._abort = False

        # reset next injection schedule
        now = _Time.now().timestamp()
        self._topup_next = now - (now % (24*60*60)) + 3*60*60
        self.run_callbacks('TopUpNextInj-Mon', self._topup_next)

    def _do_topup(self):
        # update bucket list according to settings
        self._update_bucket_list()

        # update next injection schedule
        now, period = _Time.now().timestamp(), self._topup_period
        self._topup_next = now - (now % period) + period
        self.run_callbacks('TopUpNextInj-Mon', self._topup_next)

        # prepare subsystems state
        self._prepare_topup()

        self._bias_feedback.do_update_models = True

        while self._mode == _Const.InjMode.TopUp:
            if not self._check_allok_2_inject():
                break

            self._update_topupsts(_Const.TopUpSts.Waiting)
            self._update_log('Waiting for next injection...')
            if not self._wait_topup_period():
                break
            self._bias_feedback.already_set = False

            self._update_log('Top-up period elapsed. Preparing...')
            if not self.currinfo_dev.connected:
                self._update_log('ERR:CurrInfo device disconnected.')
                break
            if self.currinfo_dev.current < self._target_current * 1.02:
                self._update_topupsts(_Const.TopUpSts.TurningOn)
                self._update_log('Starting injection...')
                if not self._start_injection():
                    break

                self._update_topupsts(_Const.TopUpSts.Injecting)
                self._update_log('Injecting...')
                if not self._wait_injection():
                    break

                self._update_bucket_list()
            else:
                self._update_topupsts(_Const.TopUpSts.Skipping)
                self._update_log('Skipping injection...')
                _time.sleep(2)

            if self._topup_pustandbyenbl:
                self._handle_aspu_standby_state(_Const.StandbyInject.Standby)
            if self._topup_liwarmupenbl:
                self._handle_liti_warmup_state(_Const.StandbyInject.Standby)
            if self._topup_bopsstandbyenbl:
                self._handle_bops_standby_state(_Const.StandbyInject.Standby)
            if self._topup_borfstandbyenbl:
                self._handle_borf_standby_state(_Const.StandbyInject.Standby)

            self._topup_next += self._topup_period
            self.run_callbacks('TopUpNextInj-Mon', self._topup_next)

        self._handle_aspu_standby_state(_Const.StandbyInject.Inject)
        self._handle_liti_warmup_state(_Const.StandbyInject.Standby)
        self._handle_bops_standby_state(_Const.StandbyInject.Inject)
        self._handle_borf_standby_state(_Const.StandbyInject.Inject)

        self._bias_feedback.do_update_models = False

        # update top-up status
        self._update_topupsts(_Const.TopUpSts.Off)
        self._update_log('Stopped top-up loop.')
        if not self._abort or self._setting_mode:
            self.run_callbacks('TopUpState-Sel', _Const.OffOn.Off)

    def _wait_topup_period(self):
        while _time.time() < self._topup_next:
            if not self._check_allok_2_inject(show_warn=False):
                return False
            _time.sleep(1)

            remaining = round(self._topup_next - _time.time())
            text = 'Remaining time: {}s'.format(remaining)
            self.run_callbacks('Log-Mon', text)
            if remaining % 60 == 0:
                _log.info(text)

            # prepare subsystems
            if remaining <= self._topup_puwarmuptime:
                self._handle_aspu_standby_state(_Const.StandbyInject.Inject)
            if remaining <= self._topup_liwarmuptime:
                self._handle_liti_warmup_state(_Const.StandbyInject.Inject)
            if remaining <= self._topup_bopswarmuptime:
                self._handle_bops_standby_state(_Const.StandbyInject.Inject)
            if remaining <= self._topup_borfwarmuptime:
                self._handle_borf_standby_state(_Const.StandbyInject.Inject)

            # bias fb
            cond = remaining <= _Const.BIASFB_AHEADSETIME
            cond &= bool(self._bias_feedback.loop_state)
            cond &= not self._bias_feedback.already_set
            if cond and self.currinfo_dev.connected:
                dcur = self._bias_feedback.get_delta_current_per_pulse(
                    per=self._topup_period, nrpul=self._topup_nrpulses,
                    curr_avg=self._target_current,
                    curr_now=self.currinfo_dev.current,
                    ltime=self.currinfo_dev.lifetime)
                self._update_log(f'BiasFB required InjCurr: {dcur:.3f}mA')
                bias = self._bias_feedback.get_bias_voltage(dcur)
                self.run_callbacks('MultBunBiasVolt-SP', bias)
                self.set_multbunbiasvolt(bias)
                self._bias_feedback.already_set = True

            if _time.time() >= self._topup_next - self._topup_headstarttime:
                return True

        self._update_log('Remaining time: 0s')
        return True

    def _prepare_topup(self):
        # If remaining time is too short do not put in standby or warmup
        # PU
        standby = _Const.StandbyInject.Standby
        if self._topup_pustandbyenbl and \
                self._topup_next - _time.time() > self._topup_puwarmuptime*2:
            self._handle_aspu_standby_state(standby)

        # LI
        if self._topup_liwarmupenbl and \
                self._topup_next - _time.time() > self._topup_liwarmuptime*2:
            self._handle_liti_warmup_state(standby)

        # BO PS
        if self._topup_bopsstandbyenbl and \
                self._topup_next - _time.time() > self._topup_bopswarmuptime*2:
            self._handle_bops_standby_state(standby)

        # BO RF
        if self._topup_borfstandbyenbl and \
                self._topup_next - _time.time() > self._topup_borfwarmuptime*2:
            self._handle_borf_standby_state(standby)

    def _handle_aspu_standby_state(self, state):
        if self._aspu_standby_state == state:
            return
        self._aspu_standby_state = state

        factor = 1 if state == _Const.StandbyInject.Inject else 0.5
        self._update_log(f'Setting PU Voltage to {factor*100}%...')
        for idx, dev in enumerate(self._pu_devs):
            if not dev.connected:
                self._update_log('WARN:'+dev.devname+' disconnected.')
                continue
            dev.voltage = self._pu_refvolt[idx] * factor
        self._update_log('...done.')

    def _handle_liti_warmup_state(self, state):
        if self._liti_warmup_state == state:
            return
        self._liti_warmup_state = state

        event = 'RmpBO' if state == _Const.StandbyInject.Inject else 'Linac'
        self._hlti_dev.change_triggers_source(
            self._li_trig_names, new_src=event, printlog=False)
        self._update_log('LI timing configured.')

    def _handle_bops_standby_state(self, state):
        if self._bops_standby_state == state:
            return
        self._bops_standby_state = state

        trigstate = int(state == _Const.StandbyInject.Inject)
        for trig in self._bops_trig_devs:
            trig.state = trigstate
        self._update_log('BO PS timing configured.')

    def _handle_borf_standby_state(self, state):
        if self._borf_standby_state == state:
            return
        self._borf_standby_state = state

        trigstate = int(state == _Const.StandbyInject.Inject)
        for trig in self._borf_trig_devs:
            trig.state = trigstate
        self._update_log('BO RF timing configured.')

    # --- auxiliary log methods ---

    def _launch_watch_dev_thread(self):
        if self._thread_watdev is None or not self._thread_watdev.is_alive():
            self._thread_watdev = _epics.ca.CAThread(
                target=self._watch_dev_process, daemon=True)
            self._thread_watdev.start()

    def _watch_dev_process(self):
        running = True
        while running:
            running = True
            for process in self._p2w:
                watcher = self._p2w[process]['watcher']
                status = self._p2w[process]['status']
                watcher_running = watcher is not None and watcher.is_alive()
                if status and not watcher_running:
                    newsts = _Const.IdleRunning.Idle
                    self._p2w[process]['status'] = newsts
                    self.run_callbacks(process+'CmdSts-Mon', newsts)
                running |= watcher_running
            _time.sleep(0.1)

    def _update_dev_status(self, sts, **kws):
        _ = kws
        if 'err:' in sts.lower():
            sts = sts.split(':')[1]
            msgs = ['ERR:'+sts[i:i+35] for i in range(0, len(sts), 35)]
        else:
            msgs = [sts[i:i+39] for i in range(0, len(sts), 39)]
        for msg in msgs:
            self._update_log(msg)

    def _update_log(self, msg):
        if 'ERR' in msg:
            _log.error(msg[4:])
        elif 'FATAL' in msg:
            _log.error(msg[6:])
        elif 'WARN' in msg:
            _log.warning(msg[5:])
        else:
            _log.info(msg)
        self.run_callbacks('Log-Mon', msg)

    def _update_topupsts(self, sts):
        self._topup_state_sts = sts
        self.run_callbacks('TopUpState-Sts', sts)

    # --- auxiliary update status methods ---

    def _update_diagstatus(self):
        """Run as a thread scanning PVs."""
        # constants of alarms to ignore
        psdiag = _PSDiagEnum.DIAG_STATUS_LABELS_AS
        psalrm = int(1 << len(psdiag)) - 1 - int(1 << psdiag.index('Alarms'))
        lipudiag = _LIDiagEnum.DIAG_STATUS_LABELS_PU
        lipualrm = int(1 << len(lipudiag)) - 1
        lipualrm -= int(1 << lipudiag.index('TRIG_Norm not ok'))
        lipualrm -= int(1 << lipudiag.index('Pulse_Current not ok'))

        # scan
        tplanned = 1.0/App.SCAN_FREQUENCY
        while not self.quit:
            if not self.scanning:
                _time.sleep(tplanned)
                continue

            _t0 = _time.time()

            # update sections status
            status_problems = set()
            for sec, sub2pvs in self._pvs_diag.items():
                status_sec = 0
                lbls = _get_sts_lbls(sec)
                for sub, d2pv in sub2pvs.items():
                    if sub not in lbls:
                        continue
                    bit = lbls.index(sub)
                    problems = set()
                    for pvo in d2pv.values():
                        if pvo.connected:
                            value = pvo.value
                            # disregard alarms
                            if sub == 'PS':
                                value = _np.bitwise_and(int(value), psalrm)
                            elif self._mode != _Const.InjMode.Decay and \
                                    sec == 'LI' and sub == 'PU':
                                value = _np.bitwise_and(int(value), lipualrm)
                            nok = value > 0
                        else:
                            nok = True
                        if nok:
                            problems.add(_PVName(pvo.pvname).device_name)
                    val = 1 if problems else 0
                    status_sec = _updt_bit(status_sec, bit, val)
                    if len(problems) > 1:
                        status_problems.add(sec+' '+sub)
                    else:
                        status_problems.update(problems)
                self._status[sec] = status_sec
                self.run_callbacks('DiagStatus'+sec+'-Mon', status_sec)
            self._status_problems = status_problems

            # compile general status
            lbls = _get_sts_lbls()
            status_all = 0
            for bit, sec in enumerate(lbls):
                val = self._status[sec] != 0
                status_all = _updt_bit(status_all, bit, val)
            self._status['All'] = status_all
            self.run_callbacks('DiagStatus-Mon', status_all)

            ttook = _time.time() - _t0
            tsleep = tplanned - ttook
            if tsleep > 0:
                _time.sleep(tsleep)
            else:
                _log.warning(
                    'DiagStatus check took more than planned... '
                    '{0:.3f}/{1:.3f} s'.format(ttook, tplanned))

    def _update_injstatus(self):
        tplanned = 1.0/App.SCAN_FREQUENCY
        while not self.quit:
            if not self.scanning:
                _time.sleep(tplanned)
                continue

            _t0 = _time.time()

            value = 0

            # TI ContinuousEvt is off
            val = 1 if not self._evg_dev.connected else \
                self._evg_dev.continuous_state != 1
            value = _updt_bit(value, 0, val)

            # BucketList not synced
            if self._mode == _Const.InjMode.Decay:
                val = 1 if not self._evg_dev.connected else \
                    self._evg_dev.bucketlist_sync != 1
                value = _updt_bit(value, 1, val)

            if self.egun_dev.connected:
                # EGBiasPS voltage diff. from desired
                volt = self._sglbunbiasvolt \
                    if self._type == _Const.InjType.SingleBunch \
                    else self._multbunbiasvolt
                val = abs(self.egun_dev.bias.voltage_mon - volt) > \
                    self.egun_dev.bias_voltage_tol
                value = _updt_bit(value, 2, val)

                # EGFilaPS current diff. from nominal
                val = not self.egun_dev.is_fila_on
                value = _updt_bit(value, 3, val)

                # EGHVPS voltage diff. from nominal
                val = not self.egun_dev.is_hv_on
                value = _updt_bit(value, 4, val)

                # EGPulsePS setup is diff. from desired
                val = _ETypes.INJTYPE[self._type] != \
                    _ETypes.INJTYPE_MON[self._type_mon]
                value = _updt_bit(value, 5, val)

                # EGTriggerPS is off
                val = not self.egun_dev.trigps.is_on()
                value = _updt_bit(value, 6, val)
            else:
                value = _updt_bit(value, 2, 1)
                value = _updt_bit(value, 3, 1)
                value = _updt_bit(value, 4, 1)
                value = _updt_bit(value, 5, 1)
                value = _updt_bit(value, 6, 1)

            # Inj.System is off
            val = 1 if not self._injsys_dev.connected else \
                not self._injsys_dev.is_on
            value = _updt_bit(value, 7, val)

            self._injstatus = value
            self.run_callbacks('InjStatus-Mon', self._injstatus)

            ttook = _time.time() - _t0
            tsleep = tplanned - ttook
            if tsleep > 0:
                _time.sleep(tsleep)
            else:
                _log.warning(
                    'InjStatus check took more than planned... '
                    '{0:.3f}/{1:.3f} s'.format(ttook, tplanned))



"""Epics Devices and Device Application."""

import time as _time
import operator as _opr
import math as _math
from functools import partial as _partial

from epics.ca import ChannelAccessGetFailure as _ChannelAccessGetFailure, \
    CASeverityException as _CASeverityException
import numpy as _np

from ..envars import VACA_PREFIX as _VACA_PREFIX
from ..epics import PV as _PV, CONNECTION_TIMEOUT as _CONN_TIMEOUT, \
    GET_TIMEOUT as _GET_TIMEOUT
from ..simul import SimPV as _PVSim
from ..simul import Simulation as _Simulation
from ..namesys import SiriusPVName as _SiriusPVName

_DEF_TIMEOUT = 10  # s
_TINY_INTERVAL = 0.050  # s


class Device:
    """Epics Device.

    Parameters
    ----------
        devname: str
            Device name, to be used as PVs prefix.
        properties: (tuple, list)
            List of properties, to be used as last part of the PV names.
        auto_monitor: bool, optional
            Whether to automatically monitor PVs for changes. Used for PVs
            that do not end with '-Mon' or 'Data'. Defaults to True.
        auto_monitor_mon: bool, optional
            Whether to automatically monitor '-Mon' or 'Data' PVs for changes.
            Defaults to False (to avoid overloading the client). Set to False
            when using PV.get_timevars() to know when a PV has been updated.
    """

    CONNECTION_TIMEOUT = _CONN_TIMEOUT
    GET_TIMEOUT = _GET_TIMEOUT
    _properties = ()

    def __init__(
            self, devname, properties, auto_monitor=True,
            auto_monitor_mon=False):
        self._properties = properties[:]
        self._auto_monitor = auto_monitor
        self._auto_monitor_mon = auto_monitor_mon
        self._devname, self._pvs = self._create_pvs(devname)

    @property
    def devname(self):
        """Return device name."""
        return self._devname

    @property
    def properties(self):
        """Return device properties."""
        return self._properties

    @property
    def simulators(self):
        """Return simulator."""
        sims = set()
        for pvname in self.pvnames:
            sims.update(_Simulation.simulator_find(pvname))
        return sims

    @property
    def pvnames(self):
        """Return device PV names."""
        pvnames = {pv.pvname for pv in self._pvs.values()}
        return pvnames

    @property
    def connected(self):
        """Return PVs connection status."""
        for pvobj in self._pvs.values():
            if not pvobj.connected:
                return False
        return True

    @property
    def auto_monitor_status(self):
        """Return PVs auto_monitor statuses."""
        return {pvn: pv.auto_monitor for pvn, pv in self._pvs.items()}

    @property
    def disconnected_pvnames(self):
        """Return list of disconnected device PVs."""
        set_ = set()
        for pvobj in self._pvs.values():
            if not pvobj.connected:
                set_.add(pvobj.pvname)
        return set_

    def set_auto_monitor(self, pvname, value):
        """Set auto_monitor state of individual PVs."""
        if pvname not in self._pvs:
            return False
        pvobj = self._pvs[pvname]
        try:
            # TODO verify need of int
            pvobj.auto_monitor = int(value)
        except (_ChannelAccessGetFailure, _CASeverityException):
            # exceptions raised in a Virtual Circuit Disconnect (192)
            # event. If the PV IOC goes down, for example.
            print('Could not set auto_monitor of {}'.format(pvobj.pvname))
            return False
        return True

    def update(self):
        """Update device properties."""
        for pvobj in self._pvs.values():
            pvobj.get()

    def pv_object(self, propty):
        """Return PV object for a given device property."""
        return self._pvs[propty]

    def pv_ctrlvars(self, propty):
        """Return PV object control variable."""
        return self._pvs[propty].get_ctrlvars()

    def pv_attribute_values(self, attribute):
        """Return property-value dict of a given attribute for all PVs."""
        attributes = dict()
        for propty in self._properties:
            pvobj = self._pvs[propty]
            attributes[pvobj.pvname] = getattr(pvobj, attribute)
        return attributes

    @property
    def hosts(self):
        """Return dict of IOC hosts providing device properties."""
        return self.pv_attribute_values('host')

    @property
    def values(self):
        """Return dict of property values."""
        return self.pv_attribute_values('value')

    def wait_for_connection(self, timeout=None):
        """Wait for connection."""
        for pvobj in self._pvs.values():
            res = pvobj.wait_for_connection(timeout)
            if not res:
                return False
        return True

    def __getitem__(self, propty):
        """Return value of property."""
        pvobj = self._pvs[propty]
        try:
            value = pvobj.get(timeout=Device.GET_TIMEOUT)
        except (_ChannelAccessGetFailure, _CASeverityException):
            # exceptions raised in a Virtual Circuit Disconnect (192)
            # event. If the PV IOC goes down, for example.
            print('Could not get value of {}'.format(pvobj.pvname))
            value = None
        return value

    def __setitem__(self, propty, value):
        """Set value of property."""
        pvobj = self._pvs[propty]
        pvobj.value = value

    # --- private methods ---

    def _create_pvs(self, devname):
        if devname:
            devname = _SiriusPVName(devname)

        pvs = dict()
        for propty in self._properties:
            pvname = self._get_pvname(devname, propty)
            auto_monitor = self._auto_monitor
            if pvname.endswith(('-Mon', 'Data')):
                auto_monitor = self._auto_monitor_mon
            in_sim = _Simulation.pv_check(pvname)
            pvclass = _PVSim if in_sim else _PV
            pvs[propty] = pvclass(
                pvname, auto_monitor=auto_monitor,
                connection_timeout=Device.CONNECTION_TIMEOUT)
        return devname, pvs

    def _wait(self, propty, value, timeout=_DEF_TIMEOUT, comp='eq'):
        """."""
        def comp_(val):
            boo = comp(self[propty], val)
            if isinstance(boo, _np.ndarray):
                boo = _np.all(boo)
            return boo

        if isinstance(comp, str):
            comp = getattr(_opr, comp)

        if comp_(value):
            return True

        ntrials = int(timeout/_TINY_INTERVAL)
        for _ in range(ntrials):
            _time.sleep(_TINY_INTERVAL)
            if comp_(value):
                return True
        return False

    def _wait_float(
            self, propty, value, rel_tol=0.0, abs_tol=0.1,
            timeout=_DEF_TIMEOUT):
        """Wait until float value gets close enough of desired value."""
        func = _partial(_math.isclose, abs_tol=abs_tol, rel_tol=rel_tol)
        return self._wait(propty, value, comp=func, timeout=timeout)

    def _get_pvname(self, devname, propty):
        if devname:
            func = devname.substitute
            pvname = func(prefix=_VACA_PREFIX, propty=propty)
        else:
            pvname = _VACA_PREFIX + ('-' if _VACA_PREFIX else '') + propty
        return pvname

    def _enum_setter(self, propty, value, enums):
        if hasattr(enums, '_fields'):
            enums = enums._fields
        if value is None:
            return None
        elif isinstance(value, str) and value in enums:
            self[propty] = enums.index(value)
        elif 0 <= int(value) < len(enums):
            self[propty] = value


class ProptyDevice(Device):
    """Device with a prefix property name."""

    def __init__(self, devname, propty_prefix, properties, **kwargs):
        """."""
        self._propty_prefix = propty_prefix
        # call base class constructor
        super().__init__(devname, properties=properties, **kwargs)

    def _get_pvname(self, devname, propty):
        if devname:
            func = devname.substitute
            pvname = func(propty=self._propty_prefix + propty)
        else:
            pvname = self._propty_prefix + propty
        return pvname


# NOTE: This class is temporary. It should become deprecated once all
# devices names are in accordance with Sirius naming system
class DeviceNC(Device):
    """Non-compliant Devices.

    This device class is to be used for those devices whose
    names and PVs are not compliant to the Sirius naming system.
    """
    DEVSEP = ':'

    def _create_pvs(self, devname):
        pvs = dict()
        devname = devname or ''
        for propty in self._properties:
            pvname = devname + self.DEVSEP + propty
            auto_monitor = not pvname.endswith('-Mon')
            pvs[propty] = _PV(pvname, auto_monitor=auto_monitor)
        return devname, pvs


class DeviceApp(Device):
    """Application Device.

    This kind of device groups properties of other devices.
    """

    def __init__(self, properties, devname=None, **kwargs):
        """."""
        self._devname_app = devname

        # call base class constructor
        super().__init__(None, properties=properties, **kwargs)

    @property
    def devname(self):
        """Return application device name."""
        return self._devname_app


class Devices:
    """."""

    def __init__(self, devname, devices):
        """."""
        self._devname = devname
        self._devices = devices

        self._properties = []
        for dev in self._devices:
            if dev is not None:
                self._properties += dev.properties

    @property
    def devname(self):
        """Return device name."""
        return self._devname

    @property
    def properties(self):
        """Return device properties."""
        return self._properties

    @property
    def simulators(self):
        """Return list of simulators."""
        sims = set()
        for dev in self._devices:
            sims.update(dev.simulators)
        return sims

    @property
    def pvnames(self):
        """Return device PV names."""
        set_ = set()
        for dev in self._devices:
            set_.update(dev.pvnames)
        return set_

    @property
    def connected(self):
        """Return PVs connection status."""
        for dev in self._devices:
            if not dev.connected:
                return False
        return True

    @property
    def disconnected_pvnames(self):
        """Return list of disconnected device PVs."""
        set_ = set()
        for dev in self._devices:
            set_.update(dev.disconnected_pvnames)
        return set_

    @property
    def auto_monitor_status(self):
        """Return PVs auto_monitor statuses."""
        dic_ = dict()
        for dev in self._devices:
            dic_.update(dev.auto_monitor_status)
        return dic_

    def update(self):
        """Update device properties."""
        for dev in self._devices:
            dev.update()

    def pv_attribute_values(self, attribute):
        """Return property-value dict of a given attribute for all PVs."""
        attributes = dict()
        for dev in self._devices:
            attrs = dev.pv_attribute_values(attribute)
            attributes.update(attrs)
        return attributes

    @property
    def hosts(self):
        """Return dict of IOC hosts providing device properties."""
        return self.pv_attribute_values('host')

    @property
    def values(self):
        """Return dict of property values."""
        return self.pv_attribute_values('value')

    def wait_for_connection(self, timeout=None):
        """Wait for connection."""
        for dev in self._devices:
            if not dev.wait_for_connection(timeout=timeout):
                return False
        return True

    @property
    def devices(self):
        """Return devices."""
        return self._devices

    # --- private methods ---

    def _set_devices_propty(self, devices, propty, values, wait=0):
        """Set devices property to value(s)."""
        dev2val = self._get_dev_2_val(devices, values)
        for dev, val in dev2val.items():
            if dev.pv_object(propty).wait_for_connection():
                dev[propty] = val
                _time.sleep(wait)

    def _wait_devices_propty(
            self, devices, propty, values, comp='eq',
            timeout=_DEF_TIMEOUT, return_prob=False):
        """Wait for devices property to reach value(s)."""
        if isinstance(comp, str):
            comp = getattr(_opr, comp)
        dev2val = self._get_dev_2_val(devices, values)

        tini = _time.time()
        for _ in range(int(timeout/_TINY_INTERVAL)):
            okdevs = set()
            for k, v in dev2val.items():
                boo = comp(k[propty], v)
                if isinstance(boo, _np.ndarray):
                    boo = _np.all(boo)
                if boo:
                    okdevs.add(k)
                if _time.time() - tini > timeout:
                    break
            list(map(dev2val.__delitem__, okdevs))
            if not dev2val:
                break
            if _time.time() - tini > timeout:
                break
            _time.sleep(_TINY_INTERVAL)

        allok = not dev2val
        if return_prob:
            return allok, [dev.pv_object(propty).pvname for dev in dev2val]
        return allok

    def _get_dev_2_val(self, devices, values):
        """Get devices to values dict."""
        # always use an iterable object
        if not isinstance(devices, (tuple, list)):
            devices = [devices, ]
        # if 'values' is not iterable, consider the same value for all devices
        if not isinstance(values, (tuple, list)):
            values = len(devices)*[values]
        return {k: v for k, v in zip(devices, values)}

    def __getitem__(self, devidx):
        """Return device."""
        return self._devices[devidx]




import numpy as np
import pandas as pd
from siriuspy.clientarch.time import Time
from siriuspy.machshift.macreport import MacReport

# determine interval
time_start = Time(2022, 1, 1, 0, 0)
time_stop = Time(2022, 12, 31, 23, 59, 59)

# get data from interval
macr = MacReport()
macr.connector.timeout = 300
macr.time_start = time_start
macr.time_stop = time_stop
macr.update()

# print small report
str_ = '{:<10s}' + '{:>16s}'*9 + '{:>20s}'
print(str_.format(
    'Y-M', 'MTBF', 'MTTR',
    'Reliability', 'Progrmd hours', 'Delivd hours', 'Total hours',
    '% stable hours', 'Stable hours', 'Unstable hours', 'Current (Avg±Std)'))
str_ = '{:<10s}' + '    {:>12.3f}'*9 + '    {:4.3f} ± {:4.3f}'
print(
    str_.format(
        str(macr.time_start.year)+'-'+str(macr.time_start.month) + ' a ' +
        str(macr.time_stop.year)+'-'+str(macr.time_stop.month),
        macr.usershift_time_between_failures_average,
        macr.usershift_time_to_recover_average,
        macr.usershift_beam_reliability,
        macr.usershift_progmd_time,
        macr.usershift_delivd_time,
        macr.usershift_total_time,
        macr.usershift_relative_stable_beam_time,
        macr.usershift_total_stable_beam_time,
        macr.usershift_total_unstable_beam_time,
        macr.usershift_current_average,
        macr.usershift_current_stddev,
    ))

# get data of interest
raw_data = macr.raw_data.copy()
raw_data.pop('Shift')
raw_data.pop('EgunModes')
raw_data.pop('Distortions')
raw_data.pop('UserShiftTotal')
raw_data.pop('UserShiftStable')
failures = raw_data.pop('Failures')
raw_data['Failures - SubsystemsNOk'] = np.asarray(failures['SubsystemsNOk'], dtype=int)
raw_data['Failures - NoEBeam'] = np.asarray(failures['NoEBeam'], dtype=int)
raw_data['Failures - WrongShift'] = np.asarray(failures['WrongShift'], dtype=int)
raw_data['TimeDate'] = [Time(t).get_iso8601() for t in raw_data['Timestamp']]

# stored data in specific order
order = [
    'Timestamp', 'TimeDate', 'Current',
    'Failures - SubsystemsNOk', 'Failures - NoEBeam', 'Failures - WrongShift',
    'UserShiftProgmd', 'UserShiftInitCurr', 'UserShiftDelivd']
data = {k: raw_data[k] for k in order}

# save data to csv
df = pd.DataFrame.from_dict(data)
df.to_csv('raw_data_machine_report_2022.csv', index=False, header=True)

# check hour numbers: each point is equivalent to 5s, converting to hours
programmed = sum(data['UserShiftProgmd'])*5/60/60
delivered = sum(data['UserShiftDelivd'])*5/60/60
print('Programmed hours:', programmed)
print('Delivered hours:', delivered)



#!/usr/bin/env python-sirius
"""Fetcher module.

See https://slacmshankar.github.io/epicsarchiver_docs/userguide.html
"""

from threading import Thread as _Thread
import asyncio as _asyncio
import urllib as _urllib
import ssl as _ssl
import logging as _log
import urllib3 as _urllib3
from aiohttp import ClientSession as _ClientSession

import numpy as _np

from .. import envars as _envars
from . import exceptions as _exceptions


class ClientArchiver:
    """Archiver Data Fetcher class."""

    DEFAULT_TIMEOUT = 5.0  # [s]
    SERVER_URL = _envars.SRVURL_ARCHIVER
    ENDPOINT = '/mgmt/bpl'

    def __init__(self, server_url=None, timeout=None):
        """Initialize."""
        timeout = timeout or ClientArchiver.DEFAULT_TIMEOUT
        self.session = None
        self._timeout = timeout
        self._url = server_url or self.SERVER_URL
        self._ret = None
        # print('urllib3 InsecureRequestWarning disabled!')
        _urllib3.disable_warnings(_urllib3.exceptions.InsecureRequestWarning)

    @property
    def connected(self):
        """Connected."""
        try:
            status = _urllib.request.urlopen(
                self._url, timeout=self._timeout,
                context=_ssl.SSLContext()).status
            return status == 200
        except _urllib.error.URLError:
            return False

    @property
    def timeout(self):
        """Connection timeout."""
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        """Set connection timeout."""
        self._timeout = float(value)

    @property
    def server_url(self):
        """Return URL of the Archiver server.

        Returns:
            str: URL of the server.

        """
        return self._url

    @server_url.setter
    def server_url(self, url):
        """Set the new server URL. Logs out if needed.

        Args:
            url (str): New server URL to use.

        """
        self.logout()
        self._url = url

    def login(self, username, password):
        """Open login session."""
        headers = {"User-Agent": "Mozilla/5.0"}
        payload = {"username": username, "password": password}
        url = self._create_url(method='login')
        ret = self._run_async_event_loop(
            self._create_session,
            url, headers=headers, payload=payload, ssl=False)
        if ret is not None:
            self.session, authenticated = ret
            if authenticated:
                print('Reminder: close connection after using this '
                      'session by calling logout method!')
            else:
                self.logout()
            return authenticated
        return False

    def logout(self):
        """Close login session."""
        if self.session:
            resp = self._run_async_event_loop(self._close_session)
            self.session = None
            return resp
        return None

    def getPVsInfo(self, pvnames):
        """Get PVs Info."""
        if isinstance(pvnames, (list, tuple)):
            pvnames = ','.join(pvnames)
        url = self._create_url(method='getPVStatus', pv=pvnames)
        resp = self._make_request(url, return_json=True)
        return None if not resp else resp

    def getAllPVs(self, pvnames):
        """Get All PVs."""
        if isinstance(pvnames, (list, tuple)):
            pvnames = ','.join(pvnames)
        url = self._create_url(method='getAllPVs', pv=pvnames, limit='-1')
        resp = self._make_request(url, return_json=True)
        return None if not resp else resp

    def deletePVs(self, pvnames):
        """Delete PVs."""
        if not isinstance(pvnames, (list, tuple)):
            pvnames = (pvnames, )
        for pvname in pvnames:
            url = self._create_url(
                method='deletePV', pv=pvname, deleteData='true')
            self._make_request(url, need_login=True)

    def getPausedPVsReport(self):
        """Get Paused PVs Report."""
        url = self._create_url(method='getPausedPVsReport')
        resp = self._make_request(url, return_json=True)
        return None if not resp else resp

    def pausePVs(self, pvnames):
        """Pause PVs."""
        if not isinstance(pvnames, (list, tuple)):
            pvnames = (pvnames, )
        for pvname in pvnames:
            url = self._create_url(method='pauseArchivingPV', pv=pvname)
            self._make_request(url, need_login=True)

    def renamePV(self, oldname, newname):
        """Rename PVs."""
        url = self._create_url(method='renamePV', pv=oldname, newname=newname)
        return self._make_request(url, need_login=True)

    def resumePVs(self, pvnames):
        """Resume PVs."""
        if not isinstance(pvnames, (list, tuple)):
            pvnames = (pvnames, )
        for pvname in pvnames:
            url = self._create_url(method='resumeArchivingPV', pv=pvname)
            self._make_request(url, need_login=True)

    def getData(self, pvname, timestamp_start, timestamp_stop,
                process_type='', interval=None, stddev=None,
                get_request_url=False):
        """Get archiver data.

        pvname -- name of pv.
        timestamp_start -- timestamp of interval start
                           Example: '2019-05-23T13:32:27.570Z'
        timestamp_stop -- timestamp of interval stop
                           Example: '2019-05-23T13:32:27.570Z'
        process_type -- data processing type to use. Can be:
                     '', 'mean', 'median', 'std', 'variance',
                     'popvariance', 'kurtosis', 'skewness'
                     'mini', 'maxi', 'jitter', 'count', 'ncount',
                     'firstSample', 'lastSample', 'firstFill', 'lastFill',
                     'nth', 'ignoreflyers' or 'flyers'
        interval -- interval of the bin of data, in seconds
        stddev -- number of standard deviations.
                  argument used in processing 'ignoreflyers' and 'flyers'.
        """
        if isinstance(pvname, str):
            pvname = [pvname, ]
        if isinstance(timestamp_start, str):
            timestamp_start = [timestamp_start, ]
        if isinstance(timestamp_stop, str):
            timestamp_stop = [timestamp_stop, ]
        if not isinstance(timestamp_start, (list, tuple)) or \
                not isinstance(timestamp_stop, (list, tuple)):
            raise _exceptions.TypeError(
                "'timestampstart' and 'timestamp_stop' arguments must be "
                "timestamp strings or iterable.")

        pvname_orig = list(pvname)
        if process_type:
            process_str = process_type
            if interval is not None:
                process_str += '_' + str(int(interval))
                if 'flyers' in process_type and stddev is not None:
                    process_str += '_' + str(int(stddev))
            pvname = [process_str+'('+pvn+')' for pvn in pvname]

        if get_request_url:
            tstart = _urllib.parse.quote(timestamp_start[0])
            tstop = _urllib.parse.quote(timestamp_stop[-1])
            url = [self._create_url(
                method='getData.json', pv=pvn,
                **{'from': tstart, 'to': tstop})
                   for pvn in pvname]
            return url[0] if len(pvname) == 1 else url

        pvn2idcs = dict()
        all_urls = list()
        for i, pvn in enumerate(pvname):
            urls = []
            for tstart, tstop in zip(timestamp_start, timestamp_stop):
                urls.append(self._create_url(
                    method='getData.json', pv=pvn,
                    **{'from': _urllib.parse.quote(tstart),
                       'to': _urllib.parse.quote(tstop)}))
            ini = len(all_urls)
            all_urls.extend(urls)
            end = len(all_urls)
            pvn2idcs[pvname_orig[i]] = _np.arange(ini, end)

        resps = self._make_request(all_urls, return_json=True)
        if not resps:
            return None

        pvn2resp = dict()
        for pvn, idcs in pvn2idcs.items():
            _ts, _vs = _np.array([]), list()
            _st, _sv = _np.array([]), _np.array([])
            for idx in idcs:
                resp = resps[idx]
                if not resp:
                    continue
                data = resp[0]['data']
                _ts = _np.r_[_ts, [v['secs'] + v['nanos']/1.0e9 for v in data]]
                for val in data:
                    _vs.append(val['val'])
                _st = _np.r_[_st, [v['status'] for v in data]]
                _sv = _np.r_[_sv, [v['severity'] for v in data]]
            if not _ts.size:
                timestamp, value, status, severity = [None, None, None, None]
            else:
                _, _tsidx = _np.unique(_ts, return_index=True)
                timestamp, status, severity = \
                    _ts[_tsidx], _st[_tsidx], _sv[_tsidx]
                value = [_vs[i] for i in _tsidx]

            pvn2resp[pvn] = dict(
                timestamp=timestamp, value=value, status=status,
                severity=severity)

        if len(pvname) == 1:
            return pvn2resp[pvname_orig[0]]
        return pvn2resp

    def getPVDetails(self, pvname, get_request_url=False):
        """Get PV Details."""
        url = self._create_url(
            method='getPVDetails', pv=pvname)
        if get_request_url:
            return url
        resp = self._make_request(url, return_json=True)
        return None if not resp else resp

    # ---------- auxiliary methods ----------

    def _make_request(self, url, need_login=False, return_json=False):
        """Make request."""
        response = self._run_async_event_loop(
            self._handle_request,
            url, return_json=return_json, need_login=need_login)
        return response

    def _create_url(self, method, **kwargs):
        """Create URL."""
        url = self._url
        if method.startswith('getData.json'):
            url += '/retrieval/data'
        else:
            url += self.ENDPOINT
        url += '/' + method
        if kwargs:
            url += '?'
            url += '&'.join(['{}={}'.format(k, v) for k, v in kwargs.items()])
        return url

    # ---------- async methods ----------
    def _run_async_event_loop(self, *args, **kwargs):
        # NOTE: Run the asyncio commands in a separated Thread to isolate
        # their EventLoop from the external environment (important for class
        # to work within jupyter notebook environment).
        _thread = _Thread(
            target=self._thread_run_async_event_loop,
            daemon=True, args=args, kwargs=kwargs)
        _thread.start()
        _thread.join()
        return self._ret

    def _thread_run_async_event_loop(self, func, *args, **kwargs):
        """Get event loop."""
        close = False
        try:
            loop = _asyncio.get_event_loop()
        except RuntimeError as error:
            if 'no current event loop' in str(error):
                loop = _asyncio.new_event_loop()
                _asyncio.set_event_loop(loop)
                close = True
            else:
                raise error
        try:
            self._ret = loop.run_until_complete(func(*args, **kwargs))
        except _asyncio.TimeoutError:
            raise _exceptions.TimeoutError

        if close:
            loop.close()

    async def _handle_request(
            self, url, return_json=False, need_login=False):
        """Handle request."""
        if self.session is not None:
            response = await self._get_request_response(
                url, self.session, return_json)
        elif need_login:
            raise _exceptions.AuthenticationError('You need to login first.')
        else:
            async with _ClientSession() as sess:
                response = await self._get_request_response(
                    url, sess, return_json)
        return response

    async def _get_request_response(self, url, session, return_json):
        """Get request response."""
        try:
            if isinstance(url, list):
                response = await _asyncio.gather(
                    *[session.get(u, ssl=False, timeout=self._timeout)
                      for u in url])
                if any([not r.ok for r in response]):
                    return None
                if return_json:
                    jsons = list()
                    for res in response:
                        try:
                            data = await res.json()
                            jsons.append(data)
                        except ValueError:
                            _log.error(f'Error with URL {res.url}')
                            jsons.append(None)
                    response = jsons
            else:
                response = await session.get(
                    url, ssl=False, timeout=self._timeout)
                if not response.ok:
                    return None
                if return_json:
                    try:
                        response = await response.json()
                    except ValueError:
                        _log.error(f'Error with URL {response.url}')
                        response = None
        except _asyncio.TimeoutError as err_msg:
            raise _exceptions.TimeoutError(err_msg)
        return response

    async def _create_session(self, url, headers, payload, ssl):
        """Create session and handle login."""
        session = _ClientSession()
        async with session.post(
                url, headers=headers, data=payload, ssl=ssl,
                timeout=self._timeout) as response:
            content = await response.content.read()
            authenticated = b"authenticated" in content
        return session, authenticated

    async def _close_session(self):
        """Close session."""
        return await self.session.close()



"""Power Supply factory classes."""

import time as _time
import re as _re
from copy import deepcopy as _deepcopy

from ..search import PSSearch as _PSSearch
from ..thread import LoopQueueThread as _LoopQueueThread
from ..bsmp.serial import Channel as _Channel

from .csdev import get_ps_propty_database as _get_ps_propty_database
from .beaglebone import BeagleBone as _BeagleBone
from .siggen import SigGenFactory as _SigGenfactory

from .bsmp.factory import PSBSMPFactory as _PSBSMPFactory

from .pructrl.pru import PRU as _PRU
from .pructrl.prucontroller import PRUController as _PRUController

from .psctrl.pscreaders import Constant as _Constant
from .psctrl.pscwriters import Setpoint as _Setpoint
from .psctrl.psmodel import PSModelFactory as _PSModelFactory


class BBBFactory:
    """BeagleBone factory."""

    # regexp of constant PVs whose initializations require bsmp communications.
    _regexp_constant_bsmp_init = _re.compile('^(Param|Version).*-Cte$')

    @staticmethod
    def create(ethbridgeclnt_class, bbbname=None):
        """Return BBB object."""
        # create timestamp used in all TimestampBoot-Cte PVs
        timestamp = _time.time()

        # create PRU used for low-level communication
        pru = _PRU(ethbridgeclnt_class, bbbname=bbbname)

        # create LoopQueueThread that queue all serial operations
        prucqueue = _LoopQueueThread()

        # build dicts for grouped udc.
        psmodels, devices, freqs = \
            BBBFactory._build_udcgrouped(bbbname)

        if len(psmodels) > 1:
            # more than one psmodel under the same beagle. there will be two
            # PRUController objects pushing operation requests into the common
            # LoopQueueThread. we have to use to lock mechanism in bsmp.Channel
            # in order to avoid one PRUController process thread interpreting
            # the operation request of the other PRUController scan thread.
            _Channel.create_lock()

        # power supply controllers and databases
        # dbase: a pvname-epicsdbase dictionary containing all
        # beaglebone PVs. the epics server database is initialized
        # using this dictionary.
        controllers, databases, dbase = \
            BBBFactory._build_controllers(
                timestamp, pru, prucqueue, psmodels, devices, freqs)

        return _BeagleBone(controllers, databases), dbase

    # --- private methods ---

    @staticmethod
    def _build_controllers(
            timestamp, pru, prucqueue, psmodels, devices, freqs):

        # dbase is a pvname to epics dbase dictionary will all PVs
        dbase = dict()
        controllers = dict()  # one controller per psmodel
        databases = dict()  # controllers databases

        for psmodel_name, psmodel in psmodels.items():

            # devices of that ps model
            devs = devices[psmodel_name]

            # create pru controller for devices
            freq = freqs[psmodel_name]
            freq = None if freq == 0 else freq
            pru_controller = _PRUController(pru, prucqueue,
                                            psmodel, devs,
                                            processing=False,
                                            scanning=False,
                                            freq=freq,
                                            init=False)

            devname2devid = dict()
            readers, writers = dict(), dict()

            for device in devs:

                # psname and bsmp device id
                psname, devid = device

                # get model database
                pstype = _PSSearch.conv_psname_2_pstype(psname)
                database = _get_ps_propty_database(psmodel_name, pstype)

                # build setpoints
                setpoints = BBBFactory._build_setpoints(
                    (device, ), database)

                # build readers and writers and add them to dicts
                _readers, _writers = BBBFactory._build_readers_writers(
                    timestamp, dbase, psmodel, setpoints, (device, ), database,
                    pru_controller)
                _ = readers.update(_readers), writers.update(_writers)

                # build device_ids dict
                devname2devid[psname] = devid

                # add database to dictionary
                databases[psname] = database

            # build controller
            controller = psmodel.controller(
                readers, writers, pru_controller, devname2devid)

            # add controller to dictionary
            for device in devs:
                psname, devid = device
                controllers[psname] = controller

        return controllers, databases, dbase

    @staticmethod
    def _build_udcgrouped(bbbname):

        # get names of all udcs under a beaglebone
        udcs = _PSSearch.conv_bbbname_2_udc(bbbname)

        psmodels, devices, freqs = dict(), dict(), dict()
        for udc in udcs:

            # add udc devices
            devs = _PSSearch.conv_udc_2_bsmps(udc)
            print('UDC: ', udc, ', DEVICES: ', devs)
            psmodel_name = BBBFactory._check_psmodels(devs)
            if psmodel_name in psmodels:
                devices[psmodel_name].extend(devs)
            else:
                devices[psmodel_name] = devs

            # add udc psmodel
            if psmodel_name not in psmodels:
                psmodel = _PSModelFactory.create(psmodel_name)
                psmodels[psmodel_name] = psmodel

            # add udc/bbb freqs
            try:
                freq = _PSSearch.conv_bbbname_2_freq(udc)
            except KeyError:
                try:
                    freq = _PSSearch.conv_bbbname_2_freq(bbbname)
                except KeyError:
                    freq = 0.0
            if psmodel_name not in freqs:
                freqs[psmodel_name] = freq
            else:
                freqs[psmodel_name] = max(freqs[psmodel_name], freq)

        return psmodels, devices, freqs

    @staticmethod
    def _build_setpoints(devices, database):
        # Build pvname-setpoints dictionary
        setpoints = dict()
        for field in database:
            for device in devices:
                devname = device[0]
                if _Setpoint.match(field):
                    setpoints[devname + ':' + field] = \
                        _Setpoint(field, database[field])
        return setpoints

    @staticmethod
    def _build_readers_writers(
            timestamp, dbase, model, setpoints, devices, database,
            pru_controller):

        readers, writers = dict(), dict()

        for field in database:
            if _Setpoint.match(field):
                # writers for setpoint field
                field_writers = \
                    BBBFactory._get_writers(
                        model, field, devices, setpoints, pru_controller)
                writers.update(field_writers)
                # corresponding readers
                for devname, devid in devices:
                    pvname = devname + ':' + field
                    dbase[pvname] = _deepcopy(database[field])
                    readers[pvname] = setpoints[pvname]
            elif _Constant.match(field) and \
                    not BBBFactory._regexp_constant_bsmp_init.match(field):
                # readers for const fields whose initializations
                # do not require bsmp communication
                for devname, devid in devices:
                    if field == 'TimestampBoot-Cte':
                        # update bootime in epics database with timestamp
                        database[field]['value'] = timestamp
                    pvname = devname + ':' + field
                    readers[pvname] = _Constant(database[field]['value'])
                    dbase[pvname] = _deepcopy(database[field])
            else:
                # readers for other fields
                for devname, devid in devices:
                    pvname = devname + ':' + field
                    dbase[pvname] = _deepcopy(database[field])
                    readers[pvname] = model.reader(
                        devid, field, pru_controller)
        return readers, writers

    @staticmethod
    def _check_psmodels(devices):
        """Check number of ps models.

        Raise exception in case the given devices have more than on psmodel
        type.
        """
        name2model = _PSSearch.conv_psname_2_psmodel
        psmodels = {name2model(psname) for psname, bsmp_id in devices}
        if len(psmodels) > 1:
            raise ValueError('Different psmodels in the same UDC')
        return psmodels.pop()

    @staticmethod
    def _get_writers(
            model, field, devices, setpoints, pru_controller):
        # NOTE: Each pwrsupply should have all variables independent
        #       in the near future.
        # if field in ('CycleType-Sel', 'CycleNrCycles-SP',
        #              'CycleFreq-SP', 'CycleAmpl-SP',
        #              'CycleOffset-SP', 'CycleAuxParam-SP'):
        #     # Make one object for all devices (UDC-shared)
        #     ids, sps = list(), list()
        #     for devname, devid in devices:
        #         pvname = devname + ':' + field
        #         ids.append(devid)
        #         sps.append(setpoints[pvname])
        #     writer = model.writer(
        #         ids, field, pru_controller, _Setpoints(sps))
        #     return {device[0] + ':' + field: writer
        #             for device in devices}
        writers = dict()
        for devname, devid in devices:
            setpoint = setpoints[devname + ':' + field]
            writers[devname + ':' + field] = model.writer(
                [devid], field, pru_controller, setpoint)
        return writers


class PSBSMPFactory:
    """Power supply BSMP factory.

    Class methods create return BSMP objects whose methods can be used to
    communicate with a given power supply,
    """

    @staticmethod
    def create(psname, pru=None):
        """."""
        udc = _PSSearch.conv_bbbname_2_udc(psname)
        psnames, devids = zip(*_PSSearch.conv_udc_2_bsmps(udc))
        idx = psnames.index(psname)

        # device ID and PS model
        devid = devids[idx]
        psmodel = _PSSearch.conv_psname_2_psmodel(psname)

        # pru
        if pru is None:
            bbbname = _PSSearch.conv_psname_2_bbbname(psname)
            pru = _PRU(bbbname)

        psbsmp = _PSBSMPFactory.create(
            psmodel=psmodel, slave_address=devid, pru=pru)

        return psbsmp


# Expose SigGenFactory class here
SigGenFactory = _SigGenfactory



"""Test."""

from matplotlib import pyplot as plt
import numpy as np
from siriuspy.clientarch.time import Time
from siriuspy.machshift.macreport import MacReport

# failure statistics per month
intervals = [
    [Time(2021, 1, 1, 0, 0), Time(2021, 1, 31, 23, 59, 59)],
    [Time(2021, 2, 1, 0, 0), Time(2021, 2, 28, 23, 59, 59)],
    [Time(2021, 3, 1, 0, 0), Time(2021, 3, 31, 23, 59, 59)],
    [Time(2021, 4, 1, 0, 0), Time(2021, 4, 30, 23, 59, 59)],
    [Time(2021, 5, 1, 0, 0), Time(2021, 5, 31, 23, 59, 59)],
    [Time(2021, 6, 1, 0, 0), Time(2021, 6, 30, 23, 59, 59)],
    [Time(2021, 7, 1, 0, 0), Time(2021, 7, 31, 23, 59, 59)],
    [Time(2021, 8, 1, 0, 0), Time(2021, 8, 31, 23, 59, 59)],
    [Time(2021, 9, 1, 0, 0), Time(2021, 9, 30, 23, 59, 59)],
    [Time(2021, 10, 1, 0, 0), Time(2021, 10, 31, 23, 59, 59)],
    [Time(2021, 11, 1, 0, 0), Time(2021, 11, 30, 23, 59, 59)],
    [Time(2021, 12, 1, 0, 0), Time(2021, 12, 31, 23, 59, 59)],
    [Time(2022, 1, 1, 0, 0), Time(2022, 1, 31, 23, 59, 59)],
    [Time(2022, 2, 1, 0, 0), Time(2022, 2, 28, 23, 59, 59)],
    [Time(2022, 3, 1, 0, 0), Time(2022, 3, 31, 23, 59, 59)],
    [Time(2022, 4, 1, 0, 0), Time(2022, 4, 30, 23, 59, 59)],
    [Time(2022, 5, 1, 0, 0), Time(2022, 5, 31, 23, 59, 59)],
    [Time(2022, 6, 1, 0, 0), Time(2022, 6, 30, 23, 59, 59)],
    [Time(2022, 7, 1, 0, 0), Time(2022, 7, 31, 23, 59, 59)],
    [Time(2022, 8, 1, 0, 0), Time(2022, 8, 31, 23, 59, 59)],
    [Time(2022, 9, 1, 0, 0), Time(2022, 9, 30, 23, 59, 59)],
    [Time(2022, 10, 1, 0, 0), Time(2022, 10, 31, 23, 59, 59)],
    [Time(2022, 11, 1, 0, 0), Time(2022, 11, 30, 23, 59, 59)],
    [Time(2022, 12, 1, 0, 0), Time(2022, 12, 31, 23, 59, 59)],
    [Time(2023, 1, 1, 0, 0), Time(2023, 1, 31, 23, 59, 59)],
    [Time(2023, 2, 1, 0, 0), Time(2023, 2, 28, 23, 59, 59)],
    [Time(2023, 3, 1, 0, 0), Time(2023, 3, 31, 23, 59, 59)],
]

macreports = dict()
for intvl in intervals:
    macreports[intvl[0]] = MacReport()
    macreports[intvl[0]].connector.timeout = 30
    macreports[intvl[0]].time_start = intvl[0]
    macreports[intvl[0]].time_stop = intvl[1]
    macreports[intvl[0]].update()

mtbfs, mttrs, reliabs = dict(), dict(), dict()
progrmd, delivd, usertot = dict(), dict(), dict()
currmean, currstd = dict(), dict()
stable, unstable, relstable = dict(), dict(), dict()
for date, macr in macreports.items():
    mtbfs[date] = macr.usershift_time_between_failures_average
    mttrs[date] = macr.usershift_time_to_recover_average
    reliabs[date] = macr.usershift_beam_reliability
    progrmd[date] = macr.usershift_progmd_time
    delivd[date] = macr.usershift_delivd_time
    usertot[date] = macr.usershift_total_time
    currmean[date] = macr.usershift_current_average
    currstd[date] = macr.usershift_current_stddev
    stable[date] = macr.usershift_total_stable_beam_time
    unstable[date] = macr.usershift_total_unstable_beam_time
    relstable[date] = macr.usershift_relative_stable_beam_time

str_ = '{:<10s}' + '{:>16s}'*9 + '{:>20s}'
print(str_.format(
    'Y-M', 'MTBF', 'MTTR',
    'Reliability', 'Progrmd hours', 'Delivd hours', 'Total hours',
    '% stable hours', 'Stable hours', 'Unstable hours', 'Current (Avg±Std)'))
str_ = '{:<10s}' + '    {:>12.3f}'*9 + '    {:4.3f} ± {:4.3f}'
for date in macreports:
    print(str_.format(
        str(date.year)+'-'+str(date.month),
        mtbfs[date],
        mttrs[date],
        reliabs[date],
        progrmd[date],
        delivd[date],
        usertot[date],
        relstable[date],
        stable[date],
        unstable[date],
        currmean[date], currstd[date],
    ))

fig, axs = plt.subplots(3, 1, sharex=True)
fig.set_size_inches(9, 6)
fig.subplots_adjust(top=0.96, left=0.08, bottom=0.05, right=0.96)
axs[0].xaxis.axis_date()
axs[0].plot(mtbfs.keys(), mtbfs.values(), '-b')
axs[0].plot(mtbfs.keys(), mtbfs.values(), 'ob')
axs[0].set_title('MTBF')
axs[0].grid()
axs[1].xaxis.axis_date()
axs[1].plot(mttrs.keys(), mttrs.values(), '-r')
axs[1].plot(mttrs.keys(), mttrs.values(), 'or')
axs[1].set_title('MTTR')
axs[1].grid()
axs[2].xaxis.axis_date()
axs[2].plot(reliabs.keys(), reliabs.values(), '-g')
axs[2].plot(reliabs.keys(), reliabs.values(), 'og')
axs[2].set_title('Reliability')
axs[2].grid()
fig.show()

# programmed vs. delivered hours
macr = MacReport()
macr.connector.timeout = 300
macr.time_start = Time(2022, 4, 1, 0, 0)
macr.time_stop = Time(2023, 3, 31, 23, 59, 59)
macr.update()

str_ = '{:<10s}' + '{:>16s}'*9 + '{:>20s}'
print(str_.format(
    'Y-M', 'MTBF', 'MTTR',
    'Reliability', 'Progrmd hours', 'Delivd hours', 'Total hours',
    '% stable hours', 'Stable hours', 'Unstable hours', 'Current (Avg±Std)'))
str_ = '{:<10s}' + '    {:>12.3f}'*9 + '    {:4.3f} ± {:4.3f}'
print(
    str_.format(
        str(macr.time_start.year)+'-'+str(macr.time_start.month) + ' a ' +
        str(macr.time_stop.year)+'-'+str(macr.time_stop.month),
        macr.usershift_time_between_failures_average,
        macr.usershift_time_to_recover_average,
        macr.usershift_beam_reliability,
        macr.usershift_progmd_time,
        macr.usershift_delivd_time,
        macr.usershift_total_time,
        macr.usershift_relative_stable_beam_time,
        macr.usershift_total_stable_beam_time,
        macr.usershift_total_unstable_beam_time,
        macr.usershift_current_average,
        macr.usershift_current_stddev,
    ))

rd = macr.raw_data
dtimes = np.diff(rd['Timestamp'])
dtimes = np.r_[dtimes, dtimes[-1]]
dtimes = dtimes/60/60

dates = [Time(d) for d in rd['Timestamp']]

dtimes_users_progmd = dtimes*rd['UserShiftProgmd']
cum_progmd = np.cumsum(dtimes_users_progmd)
dtimes_users_impltd = dtimes*rd['UserShiftDelivd']
cum_deliv = np.cumsum(dtimes_users_impltd)

fig = plt.figure()
axs = plt.gca()
axs.xaxis.axis_date()
axs.plot(dates, cum_progmd, '-', label='Programmed')
axs.plot(dates, cum_deliv, '-', label='Delivered')
axs.grid()
axs.set_ylabel('Integrated Hours')
plt.legend(loc=4)
plt.title('Integrated User Hours')
fig.show()

new_dates, new_dtimes_progmd, new_dtimes_deliv = [], [], []
for i, d in enumerate(dates):
    if not new_dates or d.date() != new_dates[-1].date():
        new_dates.append(d)
        new_dtimes_progmd.append(dtimes_users_progmd[i])
        new_dtimes_deliv.append(dtimes_users_impltd[i])
    else:
        new_dtimes_progmd[-1] += dtimes_users_progmd[i]
        new_dtimes_deliv[-1] += dtimes_users_impltd[i]
new_cum_progmd = np.cumsum(new_dtimes_progmd)
new_cum_deliv = np.cumsum(new_dtimes_deliv)

fig = plt.figure()
axs = plt.gca()
axs.xaxis.axis_date()
axs.plot(new_dates, new_cum_progmd, '-', label='Programmed')
axs.plot(new_dates, new_cum_deliv, '-', label='Delivered')
axs.grid()
axs.set_ylabel('Integrated Hours')
plt.legend(loc=4)
plt.title('Integrated User Hours')
fig.show()

np.savetxt(
    'integrated_user_hours_2021.txt',
    np.c_[[d.timestamp() for d in new_dates],
          new_cum_progmd, new_cum_deliv])



"""LI PU device."""

import time as _time

from .device import DeviceNC as _DeviceNC
from ..csdev import Const as _Const


class LIModltr(_DeviceNC):
    """LI LLRF."""

    DEF_TIMEOUT = 10  # [s]

    class DEVICES:
        """Devices names."""

        LI_MOD1 = 'LI-01:PU-Modltr-1'
        LI_MOD2 = 'LI-01:PU-Modltr-2'
        ALL = (LI_MOD1, LI_MOD2)

    _properties = (
        'Remote', 'OutPut_Status', 'RESET',
        'Run/Stop', 'PreHeat', 'Charge_Allowed', 'TrigOut_Allowed',
        'Emer_Stop', 'CPS_ALL', 'Thy_Heat', 'Kly_Heat', 'LV_Rdy_OK',
        'HV_Rdy_OK', 'TRIG_Rdy_OK', 'MOD_Self_Fault', 'MOD_Sys_Ready',
        'TRIG_Norm', 'Pulse_Current',
        'RUN_STOP', 'PREHEAT', 'CHARGE', 'TRIGOUT',
        'WRITE_I', 'READI', 'WRITE_V', 'READV')

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in LIModltr.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=LIModltr._properties)

    @property
    def run_stop(self):
        """RUN_STOP."""
        return self['RUN_STOP']

    @property
    def preheat(self):
        """PREHEAT."""
        return self['PREHEAT']

    @property
    def charge(self):
        """CHARGE."""
        return self['CHARGE']

    @charge.setter
    def charge(self, value):
        self._enum_setter('CHARGE', value, _Const.OffOn)

    @property
    def trigout(self):
        """TRIGOUT."""
        return self['TRIGOUT']

    @trigout.setter
    def trigout(self, value):
        self._enum_setter('TRIGOUT', value, _Const.OffOn)

    @property
    def voltage(self):
        """READV."""
        return self['READV']

    @voltage.setter
    def voltage(self, value):
        """WRITE_V."""
        self['WRITE_V'] = value

    @property
    def current(self):
        """READI."""
        return self['READI']

    @current.setter
    def current(self, value):
        """WRITE_I."""
        self['WRITE_I'] = value

    def cmd_reset(self, timeout=DEF_TIMEOUT):
        """Reset."""
        self['RESET'] = 1
        if not self._wait('RESET', 1, timeout/2):
            return False
        _time.sleep(1)
        self['RESET'] = 0
        return self._wait('RESET', 0, timeout/2)
